from __future__ import annotations

import asyncio
import functools
import inspect
import time
from collections.abc import AsyncGenerator, Awaitable, Callable
from typing import Any, Generic, TypeVar, cast

from bluesky.protocols import (
    Configurable,
    Locatable,
    Location,
    Movable,
    Reading,
    Status,
    Subscribable,
)
from event_model import DataKey
from stamina import retry_context

from ._device import Device, DeviceConnector, LazyMock
from ._mock_signal_backend import MockSignalBackend
from ._protocol import AsyncReadable, AsyncStageable
from ._signal_backend import SignalBackend, SignalDatatypeT, SignalDatatypeV
from ._soft_signal_backend import SoftSignalBackend
from ._status import AsyncStatus
from ._utils import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    Callback,
    T,
    error_if_none,
)


async def _wait_for(coro: Awaitable[T], timeout: float | None, source: str) -> T:
    try:
        return await asyncio.wait_for(coro, timeout)
    except TimeoutError as exc:
        raise TimeoutError(source) from exc


def _add_timeout(func):
    @functools.wraps(func)
    async def wrapper(self: Signal, *args, **kwargs):
        return await _wait_for(func(self, *args, **kwargs), self._timeout, self.source)

    return wrapper


class SignalConnector(DeviceConnector):
    """Used for connecting signals with a given backend."""

    def __init__(self, backend: SignalBackend):
        self.backend = self._init_backend = backend

    async def connect_mock(self, device: Device, mock: LazyMock):
        self.backend = MockSignalBackend(self._init_backend, mock)

    async def connect_real(self, device: Device, timeout: float, force_reconnect: bool):
        self.backend = self._init_backend
        device.log.debug(f"Connecting to {self.backend.source(device.name, read=True)}")
        await self.backend.connect(timeout)


class _ChildrenNotAllowed(dict[str, Device]):
    def __setitem__(self, key: str, value: Device) -> None:
        raise KeyError(
            f"Cannot add Device or Signal child {key}={value} of Signal, "
            "make a subclass of Device instead"
        )


class Signal(Device, Generic[SignalDatatypeT]):
    """A Device with the concept of a value, with R, RW, W and X flavours.

    :param backend: The backend for providing Signal values.
    :param timeout: The default timeout for operations on the Signal.
    :param name: The name of the signal.
    """

    _connector: SignalConnector
    _child_devices = _ChildrenNotAllowed()  # type: ignore

    def __init__(
        self,
        backend: SignalBackend[SignalDatatypeT],
        timeout: float | None = DEFAULT_TIMEOUT,
        name: str = "",
        attempts: int = 1,
    ) -> None:
        super().__init__(name=name, connector=SignalConnector(backend))
        self._timeout = timeout
        self._attempts = attempts

    @property
    def source(self) -> str:
        """Returns the source of the signal.

        E.g. "ca://PV_PREFIX:SIGNAL", or "" if not available until connection.
        """
        return self._connector.backend.source(self.name, read=True)

    @property
    def datatype(self) -> type[SignalDatatypeT] | None:
        """Returns the datatype of the signal."""
        return self._connector.backend.datatype


SignalT = TypeVar("SignalT", bound=Signal)


class _SignalCache(Generic[SignalDatatypeT]):
    def __init__(self, backend: SignalBackend[SignalDatatypeT], signal: Signal) -> None:
        self._signal: Signal[Any] = signal
        self._staged = False
        self._listeners: set[Callback] = set()
        self._valid = asyncio.Event()
        self._reading: Reading[SignalDatatypeT] | None = None
        self.backend: SignalBackend[SignalDatatypeT] = backend
        try:
            asyncio.get_running_loop()
        except RuntimeError as exc:
            raise RuntimeError(
                "Need a running event loop to subscribe to a signal, "
                "are you trying to run subscribe outside a plan?"
            ) from exc
        signal.log.debug(f"Making subscription on source {signal.source}")
        backend.set_callback(self._callback)

    def close(self) -> None:
        self.backend.set_callback(None)
        self._signal.log.debug(f"Closing subscription on source {self._signal.source}")

    def _ensure_reading(self) -> Reading[SignalDatatypeT]:
        reading = error_if_none(self._reading, "Monitor not working")
        return reading

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        await self._valid.wait()
        return self._ensure_reading()

    async def get_value(self) -> SignalDatatypeT:
        reading: Reading[SignalDatatypeT] = await self.get_reading()
        return reading["value"]

    def _callback(self, reading: Reading[SignalDatatypeT]) -> None:
        self._signal.log.debug(
            f"Updated subscription: reading of source {self._signal.source} changed "
            f"from {self._reading} to {reading}"
        )
        self._reading = reading
        self._valid.set()
        # Copy the listeners in case one of the callbacks removes the listener
        # from the set
        for callback in list(self._listeners):
            self._notify(callback)

    def _notify(
        self,
        function: Callback[dict[str, Reading[SignalDatatypeT]]],
    ) -> None:
        function({self._signal.name: self._ensure_reading()})

    def subscribe(self, function: Callback) -> None:
        self._listeners.add(function)
        if self._valid.is_set():
            self._notify(function)

    def unsubscribe(self, function: Callback) -> bool:
        if function in self._listeners:
            self._listeners.remove(function)
        else:
            self._signal.log.warning(
                f"Unsubscribe failed for signal {self._signal.name}:"
                f" subscriber {function} was not found"
                f" in listeners list: {list(self._listeners)}"
            )
        return self._staged or bool(self._listeners)

    def set_staged(self, staged: bool) -> bool:
        self._staged = staged
        return self._staged or bool(self._listeners)


class SignalR(Signal[SignalDatatypeT], AsyncReadable, AsyncStageable, Subscribable):
    """Signal that can be read from and monitored."""

    _cache: _SignalCache | None = None

    def _backend_or_cache(
        self, cached: bool | None = None
    ) -> _SignalCache | SignalBackend:
        # If cached is None then calculate it based on whether we already have a cache
        if cached is None:
            cached = self._cache is not None
        if cached:
            cache = error_if_none(self._cache, f"{self.source} not being monitored")
            return cache
        else:
            return self._connector.backend

    def _get_cache(self) -> _SignalCache:
        if not self._cache:
            self._cache = _SignalCache(self._connector.backend, self)
        return self._cache

    def _del_cache(self, needed: bool):
        if self._cache and not needed:
            self._cache.close()
            self._cache = None

    @_add_timeout
    async def read(self, cached: bool | None = None) -> dict[str, Reading]:
        """Return a single item dict with the reading in it.

        :param cached:
            Whether to use the cached monitored value:
            - If None, use the cache if it exists.
            - If False, do an explicit get.
            - If True, explicitly use the cache and raise an error if it doesn't exist.
        """
        return {self.name: await self._backend_or_cache(cached).get_reading()}

    @_add_timeout
    async def describe(self) -> dict[str, DataKey]:
        """Return a single item dict describing the signal value."""
        return {self.name: await self._connector.backend.get_datakey(self.source)}

    @_add_timeout
    async def get_value(self, cached: bool | None = None) -> SignalDatatypeT:
        """Return the current value.

        :param cached:
            Whether to use the cached monitored value:
            - If None, use the cache if it exists.
            - If False, do an explicit get.
            - If True, explicitly use the cache and raise an error if it doesn't exist.
        """
        value = await self._backend_or_cache(cached).get_value()
        self.log.debug(f"get_value() on source {self.source} returned {value}")
        return value

    def subscribe_reading(
        self, function: Callback[dict[str, Reading[SignalDatatypeT]]]
    ) -> None:
        """Subscribe to updates in the reading.

        :param function: The callback function to call when the reading changes.
        """
        self._get_cache().subscribe(function)

    subscribe = subscribe_reading

    def clear_sub(self, function: Callback) -> None:
        """Remove a subscription passed to `subscribe_reading`.

        :param function: The callback function to remove.
        """
        self._del_cache(self._get_cache().unsubscribe(function))

    @AsyncStatus.wrap
    async def stage(self) -> None:
        """Start caching this signal."""
        self._get_cache().set_staged(True)

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        """Stop caching this signal."""
        self._del_cache(self._get_cache().set_staged(False))


class SignalW(Signal[SignalDatatypeT], Movable):
    """Signal that can be set."""

    @AsyncStatus.wrap
    async def set(
        self,
        value: SignalDatatypeT,
        timeout: CalculatableTimeout = CALCULATE_TIMEOUT,
    ) -> None:
        """Set the value and return a status saying when it's done.

        :param value: The value to set.
        :param timeout: The timeout for the set.
        """
        if timeout == CALCULATE_TIMEOUT:
            timeout = self._timeout
        source = self._connector.backend.source(self.name, read=False)
        self.log.debug(f"Putting value {value} to backend at source {source}")
        async for attempt in retry_context(
            on=asyncio.TimeoutError,
            attempts=self._attempts,
            wait_initial=0,
            wait_jitter=0,
        ):
            with attempt:
                await _wait_for(self._connector.backend.put(value), timeout, source)
        self.log.debug(f"Successfully put value {value} to backend at source {source}")


class SignalRW(SignalR[SignalDatatypeT], SignalW[SignalDatatypeT], Locatable):
    """Signal that can be both read and set."""

    @_add_timeout
    async def locate(self) -> Location:
        """Return the setpoint and readback."""
        setpoint, readback = await asyncio.gather(
            self._connector.backend.get_setpoint(), self._backend_or_cache().get_value()
        )
        return Location(setpoint=setpoint, readback=readback)


class SignalX(Signal):
    """Signal that puts the default value."""

    @AsyncStatus.wrap
    async def trigger(self, timeout: CalculatableTimeout = CALCULATE_TIMEOUT) -> None:
        """Trigger the action and return a status saying when it's done.

        :param timeout: The timeout for the trigger.
        """
        if timeout == CALCULATE_TIMEOUT:
            timeout = self._timeout
        source = self._connector.backend.source(self.name, read=False)
        self.log.debug(f"Putting default value to backend at source {source}")
        await _wait_for(self._connector.backend.put(None), timeout, source)
        self.log.debug(f"Successfully put default value to backend at source {source}")


def soft_signal_rw(
    datatype: type[SignalDatatypeT],
    initial_value: SignalDatatypeT | None = None,
    name: str = "",
    units: str | None = None,
    precision: int | None = None,
) -> SignalRW[SignalDatatypeT]:
    """Create a read-writable Signal with a [](#SoftSignalBackend).

    May pass metadata, which are propagated into describe.

    :param datatype: The datatype of the signal.
    :param initial_value: The initial value of the signal.
    :param name: The name of the signal.
    :param units: The units of the signal.
    :param precision: The precision of the signal.
    """
    backend = SoftSignalBackend(datatype, initial_value, units, precision)
    signal = SignalRW(backend=backend, name=name)
    return signal


def soft_signal_r_and_setter(
    datatype: type[SignalDatatypeT],
    initial_value: SignalDatatypeT | None = None,
    name: str = "",
    units: str | None = None,
    precision: int | None = None,
) -> tuple[SignalR[SignalDatatypeT], Callable[[SignalDatatypeT], None]]:
    """Create a read-only Signal with a [](#SoftSignalBackend).

    May pass metadata, which are propagated into describe.
    Use soft_signal_rw if you want a device that is externally modifiable.

    :param datatype: The datatype of the signal.
    :param initial_value: The initial value of the signal.
    :param name: The name of the signal.
    :param units: The units of the signal.
    :param precision: The precision of the signal.
    :return: A tuple of the created SignalR and a callable to set its value.
    """
    backend = SoftSignalBackend(datatype, initial_value, units, precision)
    signal = SignalR(backend=backend, name=name)
    return (signal, backend.set_value)


async def observe_value(
    signal: SignalR[SignalDatatypeT],
    timeout: float | None = None,
    done_status: Status | None = None,
    done_timeout: float | None = None,
) -> AsyncGenerator[SignalDatatypeT, None]:
    """Subscribe to the value of a signal so it can be iterated from.

    The first value yielded in the iterator will be the current value of the
    Signal, and subsequent updates from the control system will result in that
    value being yielded, even if it is the same as the previous value.

    :param signal:
        Call subscribe_reading on this at the start, and clear_sub on it at the end.
    :param timeout:
        If given, how long to wait for each updated value in seconds. If an
        update is not produced in this time then raise asyncio.TimeoutError.
    :param done_status:
        If this status is complete, stop observing and make the iterator return.
        If it raises an exception then this exception will be raised by the
        iterator.
    :param done_timeout:
        If given, the maximum time to watch a signal, in seconds. If the loop is
        still being watched after this length, raise asyncio.TimeoutError. This
        should be used instead of on an 'asyncio.wait_for' timeout.

    Due to a rare condition with busy signals, it is not recommended to use this
    function with asyncio.timeout, including in an `asyncio.wait_for` loop.
    Instead, this timeout should be given to the done_timeout parameter.

    :example:
    ```python
    async for value in observe_value(sig):
        do_something_with(value)
    ```
    """
    async for _, value in observe_signals_value(
        signal,
        timeout=timeout,
        done_status=done_status,
        done_timeout=done_timeout,
    ):
        yield value


def _get_iteration_timeout(
    timeout: float | None, overall_deadline: float | None
) -> float | None:
    overall_deadline = overall_deadline - time.monotonic() if overall_deadline else None
    return min([x for x in [overall_deadline, timeout] if x is not None], default=None)


async def observe_signals_value(
    *signals: SignalR[SignalDatatypeT],
    timeout: float | None = None,
    done_status: Status | None = None,
    done_timeout: float | None = None,
) -> AsyncGenerator[tuple[SignalR[SignalDatatypeT], SignalDatatypeT], None]:
    """Subscribe to a set of signals so they can be iterated from.

    The first values yielded in the iterator will be the current values of the
    Signals, and subsequent updates from the control system will result in that
    value being yielded, even if it is the same as the previous value.

    :param signals:
        Call subscribe_reading on all the signals at the start, and clear_sub on
        it at the end.
    :param timeout:
        If given, how long to wait for ANY updated value from shared queue in seconds.
        If an update is not produced in this time then raise asyncio.TimeoutError.
    :param done_status:
        If this status is complete, stop observing and make the iterator return.
        If it raises an exception then this exception will be raised by the
        iterator.
    :param done_timeout:
        If given, the maximum time to watch a signal, in seconds. If the loop is
        still being watched after this length, raise asyncio.TimeoutError. This
        should be used instead of on an `asyncio.wait_for` timeout.

    :example:
    ```python
    async for signal, value in observe_signals_values(sig1, sig2, ..):
        if signal is sig1:
            do_something_with(value)
        elif signal is sig2:
            do_something_else_with(value)
    ```
    """
    q: asyncio.Queue[tuple[SignalR[SignalDatatypeT], SignalDatatypeT] | Status] = (
        asyncio.Queue()
    )
    # dict to store signal subscription to remove it later
    cbs: dict[SignalR, Callback] = {}

    # subscribe signal to update queue and fill cbs dict
    for signal in signals:

        def queue_value(reading: dict[str, Reading[SignalDatatypeT]], signal=signal):
            value = reading[signal.name]["value"]
            q.put_nowait((signal, value))

        cbs[signal] = queue_value
        signal.subscribe_reading(queue_value)

    if done_status is not None:
        done_status.add_callback(q.put_nowait)
    overall_deadline = time.monotonic() + done_timeout if done_timeout else None
    try:
        last_item = ()
        while True:
            if overall_deadline and time.monotonic() >= overall_deadline:
                raise TimeoutError(
                    f"observe_value was still observing signals "
                    f"{[signal.source for signal in signals]} after "
                    f"timeout {done_timeout}s"
                )
            iteration_timeout = _get_iteration_timeout(timeout, overall_deadline)
            try:
                item = await asyncio.wait_for(q.get(), iteration_timeout)
            except TimeoutError as exc:
                raise TimeoutError(
                    f"Timeout Error while waiting {iteration_timeout}s to update "
                    f"{[signal.source for signal in signals]}. "
                    f"Last observed signal and value were {last_item}"
                ) from exc
            except asyncio.CancelledError as exc:
                raise asyncio.CancelledError(
                    f"Cancelled Error while waiting {iteration_timeout}s to update "
                    f"{[signal.source for signal in signals]}. "
                    f"Last observed signal and value were {last_item}"
                ) from exc
            if done_status and item is done_status:
                if exc := done_status.exception():
                    raise exc
                else:
                    break
            else:
                last_item = cast(tuple[SignalR[SignalDatatypeT], SignalDatatypeT], item)
                yield last_item
    finally:
        for signal, cb in cbs.items():
            signal.clear_sub(cb)


class _ValueChecker(Generic[SignalDatatypeT]):
    def __init__(self, matcher: Callable[[SignalDatatypeT], bool], matcher_name: str):
        self._last_value: SignalDatatypeT | None = None
        self._matcher = matcher
        self._matcher_name = matcher_name

    async def _wait_for_value(self, signal: SignalR[SignalDatatypeT]):
        async for value in observe_value(signal):
            self._last_value = value
            if self._matcher(value):
                return

    async def wait_for_value(
        self, signal: SignalR[SignalDatatypeT], timeout: float | None
    ):
        try:
            await asyncio.wait_for(self._wait_for_value(signal), timeout)
        except TimeoutError as exc:
            raise TimeoutError(
                f"{signal.name} didn't match {self._matcher_name} in {timeout}s, "
                f"last value {self._last_value!r}"
            ) from exc


async def wait_for_value(
    signal: SignalR[SignalDatatypeT],
    match: SignalDatatypeT | Callable[[SignalDatatypeT], bool],
    timeout: float | None,
) -> None:
    """Wait for a signal to have a matching value.

    :param signal:
        Call subscribe_reading on this at the start, and clear_sub on it at the
        end.
    :param match:
        If a callable, it should return True if the value matches. If not
        callable then value will be checked for equality with match.
    :param timeout: How long to wait for the value to match.

    :example:
    ```python
    await wait_for_value(device.acquiring, 1, timeout=1)
    # or
    await wait_for_value(device.num_captured, lambda v: v > 45, timeout=1)
    ```
    """
    if callable(match):
        checker = _ValueChecker(match, match.__name__)  # type: ignore
    else:
        checker = _ValueChecker(lambda v: v == match, repr(match))
    await checker.wait_for_value(signal, timeout)


async def set_and_wait_for_other_value(
    set_signal: SignalW[SignalDatatypeT],
    set_value: SignalDatatypeT,
    match_signal: SignalR[SignalDatatypeV],
    match_value: SignalDatatypeV | Callable[[SignalDatatypeV], bool],
    timeout: float = DEFAULT_TIMEOUT,
    set_timeout: float | None = None,
    wait_for_set_completion: bool = True,
) -> AsyncStatus:
    """Set a signal and monitor another signal until it has the specified value.

    This function sets a set_signal to a specified set_value and waits for
    a match_signal to have the match_value.

    :param set_signal: The signal to set.
    :param set_value: The value to set it to.
    :param match_signal: The signal to monitor.
    :param match_value:
        The value (or callable that says if the value matches) to wait for.
    :param timeout: How long to wait for the signal to have the value.
    :param set_timeout: How long to wait for the set to complete.
    :param wait_for_set_completion:
        If False then return as soon as the match_signal matches match_value. If
        True then also wait for the set operation to complete before returning.

    :seealso:
    [](#interact-with-signals)

    :example:
    To set the setpoint and wait for the readback to match:
    ```python
    await set_and_wait_for_value(device.setpoint, 1, device.readback, 1)
    ```
    """
    # Start monitoring before the set to avoid a race condition
    values_gen = observe_value(match_signal)

    # Get the initial value from the monitor to make sure we've created it
    current_value = await anext(values_gen)

    status = set_signal.set(set_value, timeout=set_timeout)

    if callable(match_value):
        matcher: Callable[[SignalDatatypeV], bool] = match_value  # type: ignore
    else:

        def matcher(value):
            return value == match_value

        matcher.__name__ = f"equals_{match_value}"

    # If the value was the same as before no need to wait for it to change
    if not matcher(current_value):

        async def _wait_for_value():
            async for value in values_gen:
                if matcher(value):
                    break

        try:
            await asyncio.wait_for(_wait_for_value(), timeout)
            if wait_for_set_completion:
                await status
        except TimeoutError as exc:
            matcher_name = getattr(matcher, "__name__", f"<{type(matcher).__name__}>")
            raise TimeoutError(
                f"{match_signal.name} value didn't match value from"
                f" {matcher_name}() in {timeout}s"
            ) from exc

    return status


async def set_and_wait_for_value(
    signal: SignalRW[SignalDatatypeT],
    value: SignalDatatypeT,
    match_value: SignalDatatypeT | Callable[[SignalDatatypeT], bool] | None = None,
    timeout: float = DEFAULT_TIMEOUT,
    set_timeout: float | None = None,
    wait_for_set_completion: bool = True,
) -> AsyncStatus:
    """Set a signal and monitor that same signal until it has the specified value.

    This function sets a set_signal to a specified set_value and waits for
    a match_signal to have the match_value.

    :param signal: The signal to set.
    :param value: The value to set it to.
    :param match_value:
        The value (or callable that says if the value matches) to wait for.
    :param timeout: How long to wait for the signal to have the value.
    :param set_timeout: How long to wait for the set to complete.
    :param wait_for_set_completion:
        If False then return as soon as the match_signal matches match_value. If
        True then also wait for the set operation to complete before returning.

    :seealso:
    [](#interact-with-signals)

    :examples:
    To set a parameter and wait for it's value to change:
    ```python
    await set_and_wait_for_value(device.parameter, 1)
    ```
    For busy record, or other Signals with pattern:
      - Set `wait=non_zero` when creating the signal
      - Set Signal and stash the Status
      - Read the same Signal to check the operation has started
      - Return the Status so calling code can wait for operation to complete
    ```python
    status = await set_and_wait_for_value(
        device.acquire, 1, wait_for_set_completion=False
    )
    # device is now acquiring
    await status
    # device has finished acquiring
    ```
    """
    if match_value is None:
        match_value = value
    return await set_and_wait_for_other_value(
        signal,
        value,
        signal,
        match_value,
        timeout,
        set_timeout,
        wait_for_set_completion,
    )


def walk_rw_signals(device: Device) -> dict[str, SignalRW[Any]]:
    """Retrieve all SignalRWs from a device.

    Stores retrieved signals with their dotted attribute paths in a dictionary. Used as
    part of saving and loading a device.

    :param device: Device to retrieve read-write signals from.
    :param path_prefix: For internal use, leave blank when calling the method.
    :return:
        A dictionary matching the string attribute path of a SignalRW with the
        signal itself.
    """
    all_devices = walk_devices(device)
    return {path: dev for path, dev in all_devices.items() if type(dev) is SignalRW}


async def walk_config_signals(
    device: Device,
) -> dict[str, SignalRW[Any]]:
    """Retrieve all configuration signals from a device.

    Stores retrieved signals with their dotted attribute paths in a dictionary. Used as
    part of saving and loading a device.

    :param device: Device to retrieve configuration signals from.
    :return:
        A dictionary matching the string attribute path of a SignalRW with the
        signal itself.
    """
    config_names: list[str] = []
    if isinstance(device, Configurable):
        configuration = device.read_configuration()
        if inspect.isawaitable(configuration):
            configuration = await configuration
        config_names = list(configuration.keys())

    all_devices = walk_devices(device)
    return {
        path: dev
        for path, dev in all_devices.items()
        if isinstance(dev, SignalRW) and dev.name in config_names
    }


class Ignore:
    """Annotation to ignore a signal when connecting a device."""

    pass


def walk_devices(device: Device, path_prefix: str = "") -> dict[str, Device]:
    """Recursively retrieve all Devices from a device tree.

    :param device: Root device to start from.
    :param path_prefix: For internal use, leave blank when calling the method.
    :return: A dictionary mapping dotted attribute paths to Device instances.
    """
    devices: dict[str, Device] = {}
    for attr_name, attr in device.children():
        dot_path = f"{path_prefix}{attr_name}"
        devices[dot_path] = attr
        devices.update(walk_devices(attr, path_prefix=dot_path + "."))
    return devices


def walk_signal_sources(device: Device) -> dict[str, str]:
    """Recursively gather the `source` field from every Signal in a device tree.

    :param device: Root device to start from.
    :param path_prefix: For internal use, leave blank when calling the method.
    :return: A dictionary mapping dotted attribute paths to Signal source strings.
    """
    all_devices = walk_devices(device)
    return {
        path: dev.source for path, dev in all_devices.items() if isinstance(dev, Signal)
    }


class SignalDict(dict[SignalR, Any]):
    def __getitem__(self, key: SignalR[SignalDatatypeT]) -> SignalDatatypeT:
        return super().__getitem__(key)

    def __setitem__(self, key: SignalR[SignalDatatypeT], value: SignalDatatypeT):
        return super().__setitem__(key, value)



"""Tango Signals over Pytango."""

from __future__ import annotations

import logging

from tango import (
    AttrWriteType,
    DeviceProxy,
)
from tango.asyncio import DeviceProxy as AsyncDeviceProxy

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Signal,
    SignalDatatype,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
)

from ._tango_transport import (
    CommandProxyReadCharacter,
    TangoSignalBackend,
    get_command_character,
    get_python_type,
)
from ._utils import get_device_trl_and_attr

logger = logging.getLogger("ophyd_async")


def make_backend(
    datatype: type[SignalDatatypeT] | None,
    read_trl: str = "",
    write_trl: str = "",
) -> TangoSignalBackend:
    return TangoSignalBackend(datatype, read_trl, write_trl)


def tango_signal_rw(
    datatype: type[SignalDatatypeT],
    read_trl: str,
    write_trl: str = "",
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalRW[SignalDatatypeT]:
    """Create a `SignalRW` backed by 1 or 2 Tango Attribute/Command.

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    read_trl:
        The Attribute/Command to read and monitor
    write_trl:
        If given, use this Attribute/Command to write to, otherwise use read_trl
    timeout:
        The timeout for the read and write operations
    name:
        The name of the Signal

    """
    backend = make_backend(datatype, read_trl, write_trl or read_trl)
    return SignalRW(backend, timeout=timeout, name=name)


def tango_signal_r(
    datatype: type[SignalDatatypeT],
    read_trl: str,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalR[SignalDatatypeT]:
    """Create a `SignalR` backed by 1 Tango Attribute/Command.

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    read_trl:
        The Attribute/Command to read and monitor
    timeout:
        The timeout for the read operation
    name:
        The name of the Signal

    """
    backend = make_backend(datatype, read_trl, read_trl)
    return SignalR(backend, timeout=timeout, name=name)


def tango_signal_w(
    datatype: type[SignalDatatypeT],
    write_trl: str,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalW[SignalDatatypeT]:
    """Create a `SignalW` backed by 1 Tango Attribute/Command.

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    write_trl:
        The Attribute/Command to write to
    timeout:
        The timeout for the write operation
    name:
        The name of the Signal

    """
    backend = make_backend(datatype, write_trl, write_trl)
    return SignalW(backend, timeout=timeout, name=name)


def tango_signal_x(
    write_trl: str,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalX:
    """Create a `SignalX` backed by 1 Tango Attribute/Command.

    Parameters
    ----------
    write_trl:
        The Attribute/Command to write its initial value to on execute
    timeout:
        The timeout for the command operation
    name:
        The name of the Signal

    """
    backend = make_backend(None, write_trl, write_trl)
    return SignalX(backend, timeout=timeout, name=name)


async def infer_python_type(
    trl: str = "", proxy: DeviceProxy | None = None
) -> type[SignalDatatype] | None:
    """Infers the python type from the TRL."""
    # TODO: work out if this is still needed
    device_trl, tr_name = get_device_trl_and_attr(trl)
    if proxy is None:
        dev_proxy = await AsyncDeviceProxy(device_trl)  # type: ignore
    else:
        dev_proxy = proxy

    if tr_name in dev_proxy.get_command_list():
        # A Device proxy instantiated by awaiting
        # tango.asyncio.DeviceProxy is typed the same as the sync
        # despite having awaitable methods.
        config = await dev_proxy.get_command_config(tr_name)  # type: ignore
        py_type = get_python_type(config)
    elif tr_name in dev_proxy.get_attribute_list():
        config = await dev_proxy.get_attribute_config(tr_name)  # type: ignore
        py_type = get_python_type(config)
    else:
        raise RuntimeError(f"Cannot find {tr_name} in {device_trl}")
    return py_type


async def infer_signal_type(
    trl, proxy: DeviceProxy | None = None
) -> type[Signal] | None:
    device_trl, tr_name = get_device_trl_and_attr(trl)
    if proxy is None:
        dev_proxy = await AsyncDeviceProxy(device_trl)  # type: ignore
    else:
        dev_proxy = proxy

    if tr_name not in dev_proxy.get_attribute_list():
        if tr_name not in dev_proxy.get_command_list():
            raise RuntimeError(f"Cannot find {tr_name} in {device_trl}")

    if tr_name in dev_proxy.get_attribute_list():
        config = await dev_proxy.get_attribute_config(tr_name)  # type: ignore
        if config.writable in [AttrWriteType.READ_WRITE, AttrWriteType.READ_WITH_WRITE]:
            return SignalRW
        elif config.writable == AttrWriteType.READ:
            return SignalR
        else:
            return SignalW

    if tr_name in dev_proxy.get_command_list():
        config = await dev_proxy.get_command_config(tr_name)  # type: ignore
        command_character = get_command_character(config)
        if command_character == CommandProxyReadCharacter.READ:
            return SignalR
        elif command_character == CommandProxyReadCharacter.WRITE:
            return SignalW
        elif command_character == CommandProxyReadCharacter.READ_WRITE:
            return SignalRW
        elif command_character == CommandProxyReadCharacter.EXECUTE:
            return SignalX
    raise RuntimeError(f"Unable to infer signal character for {trl}")



from bluesky.utils import MsgGenerator, plan

from ophyd_async.core import Settings
from ophyd_async.plan_stubs import apply_settings

from ._detector import HDFPanda


@plan
def apply_panda_settings(settings: Settings[HDFPanda]) -> MsgGenerator[None]:
    """Apply given settings to a panda device."""
    units, others = settings.partition(lambda signal: signal.name.endswith("_units"))
    yield from apply_settings(units)
    yield from apply_settings(others)



import asyncio
import contextlib
import re
import time
import traceback
from asyncio import CancelledError
from unittest.mock import Mock

import bluesky.plan_stubs as bps
import pytest
from bluesky.protocols import Movable, Status
from bluesky.utils import FailedStatus

from ophyd_async.core import AsyncStatus, Device, completed_status


@contextlib.contextmanager
def timing_check(expected_time: float, abs_tolerance: float = 0.1):
    """Context manager to assert execution time is approximately as expected."""
    start = time.monotonic()
    yield
    end = time.monotonic()
    assert end - start == pytest.approx(expected_time, abs=abs_tolerance)


async def test_async_status_success():
    st = AsyncStatus(asyncio.sleep(0.1))
    assert isinstance(st, Status)
    assert not st.done
    assert not st.success
    await st
    assert st.done
    assert st.success


async def test_async_status_propagates_exception(failing_coroutine):
    status = AsyncStatus(failing_coroutine())
    assert status.exception() is None

    with pytest.raises(ValueError):
        await status

    assert isinstance(status.exception(), ValueError)


async def test_async_status_propagates_cancelled_error(normal_coroutine):
    coro, is_running = normal_coroutine
    status = AsyncStatus(coro())
    assert status.exception() is None

    status.task.exception = Mock(side_effect=asyncio.CancelledError(""))
    await status

    assert isinstance(status.exception(), asyncio.CancelledError)


class SelfCancellingDevice(Device):
    @AsyncStatus.wrap
    async def set(self, value):
        await asyncio.sleep(0.1)
        raise CancelledError()


async def test_async_status_propagates_cancelled_error_with_message():
    device = SelfCancellingDevice("MY_DEVICE")

    with pytest.raises(CancelledError) as e:
        await device.set(1)

    assert re.search("CancelledError while awaiting .*MY_DEVICE", e.value.args[0])


async def test_async_status_has_no_exception_if_coroutine_successful(normal_coroutine):
    coro, is_running = normal_coroutine
    status = AsyncStatus(coro())
    assert status.exception() is None

    await status

    assert status.exception() is None


@pytest.mark.parametrize("wait_to_run", [True, False])
async def test_async_status_success_if_cancelled(normal_coroutine, wait_to_run):
    cbs = []
    coro, is_running = normal_coroutine
    status = AsyncStatus(coro())
    status.add_callback(cbs.append)
    assert status.exception() is None
    if wait_to_run:
        await is_running.wait()
    status.task.cancel()
    assert not cbs
    with pytest.raises(asyncio.CancelledError):
        await status
    assert cbs == [status]
    assert status.success is False
    assert isinstance(status.exception(), asyncio.CancelledError)


async def coroutine_to_wrap(time: float):
    await asyncio.sleep(time)


async def test_async_status_wrap() -> None:
    wrapped_coroutine = AsyncStatus.wrap(coroutine_to_wrap)
    status: AsyncStatus = wrapped_coroutine(0.01)

    await status
    assert status.success is True


async def test_async_status_initialised_with_a_task(normal_coroutine):
    coro, is_running = normal_coroutine
    normal_task = asyncio.Task(coro())
    status = AsyncStatus(normal_task)

    await status
    assert status.success is True


async def test_async_status_str_for_normal_coroutine(normal_coroutine):
    coro, is_running = normal_coroutine
    normal_task = asyncio.Task(coro())
    status = AsyncStatus(normal_task)

    for comment_chunk in ["<AsyncStatus,", "normal_coroutine", "pending>"]:
        assert comment_chunk in str(status)
    await status

    for comment_chunk in ["<AsyncStatus,", "normal_coroutine", "done>"]:
        assert comment_chunk in str(status)


async def test_async_status_str_for_failing_coroutine(failing_coroutine):
    failing_task = asyncio.Task(failing_coroutine())
    status = AsyncStatus(failing_task)

    for comment_chunk in ["<AsyncStatus,", "failing_coroutine", "pending>"]:
        assert comment_chunk in str(status)
    with pytest.raises(ValueError):
        await status

    for comment_chunk in [
        "<AsyncStatus,",
        "failing_coroutine",
        "errored:",
        "ValueError",
    ]:
        assert comment_chunk in str(status)


async def test_status_complete_before_loop_complete():
    vals = []
    # If the status completes before the loop, then if should
    # break out without erroring
    with timing_check(0.01):
        async with AsyncStatus(asyncio.sleep(0.01)):
            for i in range(2):
                vals.append(i)
                await asyncio.sleep(1)
    assert vals == [0]


async def test_loop_complete_before_status_complete():
    vals = []

    # If the loop completes before the status, then it
    # should cancel the task the status is waiting for
    async def deferred_put():
        await asyncio.sleep(1)
        vals.append("deferred_put")

    with timing_check(0.02):
        async with AsyncStatus(deferred_put()):
            for i in range(2):
                vals.append(i)
                await asyncio.sleep(0.01)
    assert vals == [0, 1]
    # Check that if we wait a while it still didn't put deferred_put onto the list
    await asyncio.sleep(0.01)
    assert vals == [0, 1]


async def test_error_raised_from_with_status():
    vals = []
    # If an error is raised from the loop then it should be raised
    with timing_check(0):
        with pytest.raises(ValueError, match="breaking out"):
            async with AsyncStatus(asyncio.sleep(1)):
                for i in range(2):
                    vals.append(i)
                    raise ValueError("breaking out")
    assert vals == [0]


async def test_status_coroutine_raises_exception(failing_coroutine):
    vals = []
    # The exception from the status task should propagate out
    with timing_check(0.01):
        with pytest.raises(ValueError):
            async with AsyncStatus(failing_coroutine()):
                for i in range(10):
                    vals.append(i)
                    await asyncio.sleep(1)
    # Loop should be cancelled after first iteration when status fails
    assert vals == [0]


async def test_status_already_complete_before_entering():
    vals = []

    # Create a status that completes immediately
    status = AsyncStatus(asyncio.sleep(0))
    await asyncio.sleep(0.01)  # Ensure it's complete
    assert status.done

    # Using it as a context manager should cancel the loop after first iteration
    # (callback is added and called synchronously, but cancellation happens at next
    # await)
    with timing_check(0.01):
        async with status:
            for i in range(10):
                vals.append(i)
                await asyncio.sleep(1)
    # The loop executes one iteration before cancellation takes effect
    assert vals == [0]


async def test_finally_block_executes_when_cancelled():
    vals = []

    async with AsyncStatus(asyncio.sleep(0.01)):
        try:
            for i in range(10):
                vals.append(i)
                await asyncio.sleep(1)
        finally:
            vals.append("cleanup")

    assert vals == [0, "cleanup"]


async def test_nested_asyncstatus_context_managers_inner_breaks():
    vals = []

    # Inner status completes first, should cancel inner loop but not outer
    with timing_check(0.02):
        async with AsyncStatus(asyncio.sleep(1)):
            vals.append("outer_start")
            async with AsyncStatus(asyncio.sleep(0.01)):
                for i in range(10):
                    vals.append(f"inner_{i}")
                    await asyncio.sleep(1)
            vals.append("outer_continue")
            await asyncio.sleep(0.01)
    assert vals == ["outer_start", "inner_0", "outer_continue"]


async def test_nested_asyncstatus_context_managers_outer_breaks():
    vals = []

    # Outer status completes first, should cancel inner context manager too
    with timing_check(0.02):
        async with AsyncStatus(asyncio.sleep(0.01)):
            vals.append("outer_start")
            async with AsyncStatus(asyncio.sleep(1)):
                for i in range(10):
                    vals.append(f"inner_{i}")
                    await asyncio.sleep(1)
            vals.append("outer_continue")
            await asyncio.sleep(0.01)
    assert vals == ["outer_start", "inner_0"]


async def test_awaiting_status_inside_its_own_context():
    vals = []

    async with AsyncStatus(asyncio.sleep(0.01)) as status:
        vals.append(0)
        await status  # Explicitly await the status
        vals.append(1)  # This should not be reached

    assert vals == [0]


async def test_no_await_points_in_loop_body():
    vals = []

    with timing_check(0.01):
        async with AsyncStatus(asyncio.sleep(0)):
            for i in range(100):
                vals.append(i)
                # No await here - tight loop
    # Without await points, the loop should complete many iterations
    # before the event loop can process the cancellation
    assert len(vals) == 100


async def test_both_status_and_loop_raise_exceptions():
    vals = []

    async def failing_status():
        await asyncio.sleep(0.05)
        raise RuntimeError("Status failed")

    # Loop raises first, so its exception should propagate
    with pytest.raises(ValueError, match="Loop failed"):
        async with AsyncStatus(failing_status()):
            vals.append(0)
            await asyncio.sleep(0.01)
            raise ValueError("Loop failed")

    assert vals == [0]


async def test_status_with_precancelled_task():
    task = asyncio.create_task(asyncio.sleep(1))
    task.cancel()

    with contextlib.suppress(asyncio.CancelledError):
        await task  # Ensure it's cancelled

    status = AsyncStatus(task)
    vals = []

    # Using a pre-cancelled status cancels the loop at the first await point
    # The CancelledError is not suppressed since the context manager didn't cause it
    with pytest.raises(asyncio.CancelledError):
        async with status:
            for i in range(10):
                vals.append(i)
                await asyncio.sleep(0.1)

    # The loop executes one iteration before cancellation takes effect
    assert vals == [0]


async def test_status_context_with_exception_in_finally():
    vals = []

    with pytest.raises(RuntimeError, match="Finally block error"):
        async with AsyncStatus(asyncio.sleep(0.01)):
            try:
                for i in range(10):
                    vals.append(i)
                    await asyncio.sleep(1)
            finally:
                vals.append("finally")
                raise RuntimeError("Finally block error")

    assert vals == [0, "finally"]


class FailingMovable(Movable, Device):
    def _fail(self):
        raise ValueError("This doesn't work")

    @AsyncStatus.wrap
    async def set(self, value):
        if value:
            return self._fail()


async def test_status_propogates_traceback_under_re(RE) -> None:
    expected_call_stack = ["wait_with_error_message", "set", "_fail"]
    d = FailingMovable()
    with pytest.raises(FailedStatus) as ctx:
        RE(bps.mv(d, 3))
    # We get "The above exception was the direct cause of the following exception:",
    # so extract that first exception traceback and check
    assert ctx.value.__cause__
    assert expected_call_stack == [
        x.name for x in traceback.extract_tb(ctx.value.__cause__.__traceback__)
    ]
    # Check we get the same from the status.exception
    status: AsyncStatus = ctx.value.args[0]
    exception = status.exception()
    assert exception
    assert expected_call_stack == [
        x.name for x in traceback.extract_tb(exception.__traceback__)
    ]


async def test_async_status_exception_timeout():
    st = AsyncStatus(asyncio.sleep(0.1))
    try:
        with pytest.raises(
            ValueError,
            match=(
                "cannot honour any timeout other than 0 in an asynchronous function"
            ),
        ):
            st.exception(timeout=1.0)
    finally:
        if not st.done:
            st.task.cancel()


@pytest.fixture
def loop():
    return asyncio.get_event_loop()


def test_asyncstatus_wraps_bare_func(loop):
    async def do_test():
        @AsyncStatus.wrap
        async def coro_status():
            await asyncio.sleep(0.01)

        st = coro_status()
        assert isinstance(st, AsyncStatus)
        await asyncio.wait_for(st.task, None)
        assert st.done

    loop.run_until_complete(do_test())


def test_asyncstatus_wraps_bare_func_with_args_kwargs(loop):
    async def do_test():
        test_result = 5

        @AsyncStatus.wrap
        async def coro_status(x: int, y: int, *, z=False):
            await asyncio.sleep(0.01)
            nonlocal test_result
            test_result = x * y if z else 0

        st = coro_status(3, 4, z=True)
        assert isinstance(st, AsyncStatus)
        await asyncio.wait_for(st.task, None)
        assert st.done
        assert test_result == 12

    loop.run_until_complete(do_test())


async def test_completed_status():
    with pytest.raises(ValueError):
        await completed_status(ValueError())
    await completed_status()


async def test_device_name_in_failure_message_asyncstatus_wrap(RE):
    device_name = "MyFailingMovable"
    d = FailingMovable(name=device_name)
    with pytest.raises(FailedStatus) as ctx:
        RE(bps.mv(d, 3))
    # FailingMovable.set is decorated with @AsyncStatus.wrap
    # undecorated methods will not print the device name
    status: AsyncStatus = ctx.value.args[0]
    assert f"device: {device_name}" in repr(status)



import asyncio
from collections.abc import Sequence

import numpy as np

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorDataLogic,
    PathProvider,
    SignalR,
    StreamableDataProvider,
    StreamResourceDataProvider,
    StreamResourceInfo,
    wait_for_value,
)

from ._io import OdinIO


class OdinDataLogic(DetectorDataLogic):
    def __init__(
        self,
        path_provider: PathProvider,
        odin: OdinIO,
        detector_bit_depth: SignalR[int],
    ):
        self.path_provider = path_provider
        self.odin = odin
        self.detector_bit_depth = detector_bit_depth

    async def prepare_unbounded(self, datakey_name: str) -> StreamableDataProvider:
        # Work out where to write
        path_info = self.path_provider(datakey_name)
        # Get the current bit depth
        datatype = f"uint{await self.detector_bit_depth.get_value()}"
        # Setup the HDF writer
        filename = f"{path_info.filename}.h5"
        await asyncio.gather(
            self.odin.fp.data_datatype.set(datatype),
            self.odin.fp.data_compression.set("BSLZ4"),
            self.odin.fp.frames.set(0),
            self.odin.fp.process_frames_per_block.set(1000),
            self.odin.fp.file_path.set(str(path_info.directory_path)),
            self.odin.mw.directory.set(str(path_info.directory_path)),
            self.odin.fp.file_prefix.set(filename),
            self.odin.mw.file_prefix.set(filename),
            self.odin.mw.acquisition_id.set(filename),
        )
        # Start writing
        await self.odin.fp.start_writing.trigger()
        await wait_for_value(self.odin.fp.writing, True, timeout=DEFAULT_TIMEOUT)
        await wait_for_value(self.odin.mw.writing, True, timeout=DEFAULT_TIMEOUT)
        # Return a provider that reflects what we have made
        data_shape = await asyncio.gather(
            self.odin.fp.data_dims_0.get_value(), self.odin.fp.data_dims_1.get_value()
        )
        resource = StreamResourceInfo(
            data_key=datakey_name,
            shape=data_shape,
            chunk_shape=(1, *data_shape),
            dtype_numpy=np.dtype(datatype).str,
            parameters={"dataset": "/data"},
        )
        return StreamResourceDataProvider(
            uri=f"{path_info.directory_uri}{filename}",
            resources=[resource],
            mimetype="application/x-hdf5",
            collections_written_signal=self.odin.fp.frames_written,
        )

    async def stop(self) -> None:
        await asyncio.gather(
            self.odin.fp.stop_writing.trigger(),
            self.odin.mw.stop.trigger(),
        )

    def get_hinted_fields(self, datakey_name: str) -> Sequence[str]:
        # The main dataset is always hinted
        return [datakey_name]



import asyncio
import os
import time
import typing
from collections.abc import Awaitable, Callable
from enum import Enum
from pathlib import Path
from typing import Any, Generic, Literal, TypeVar, get_args
from unittest.mock import patch

import bluesky.plan_stubs as bps
import numpy as np
import numpy.typing as npt
import pytest
import yaml
from aioca import Subscription, purge_channel_caches
from bluesky.protocols import Location
from event_model import Dtype, Limits, LimitsRange
from ophyd.signal import EpicsSignal

from ophyd_async.core import (
    Array1D,
    NotConnectedError,
    Signal,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    StrictEnum,
    SubsetEnum,
    Table,
    YamlSettingsProvider,
    observe_value,
    set_mock_value,
    soft_signal_r_and_setter,
)
from ophyd_async.epics.core import (
    CaSignalBackend,
    PvaSignalBackend,
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    epics_signal_w,
    epics_signal_x,
)
from ophyd_async.epics.core._util import format_datatype  # noqa: PLC2701
from ophyd_async.epics.testing import (
    EpicsTestEnum,
    EpicsTestIocAndDevices,
    EpicsTestSubsetEnum,
    EpicsTestTable,
)
from ophyd_async.plan_stubs import (
    apply_settings,
    ensure_connected,
    retrieve_settings,
    store_settings,
)
from ophyd_async.testing import MonitorQueue, assert_describe_signal

T = TypeVar("T")
Protocol = Literal["ca", "pva"]


TIMEOUT = 30.0 if os.name == "nt" else 3.0


@pytest.fixture(scope="module")
def ioc_devices():
    ioc_devices = EpicsTestIocAndDevices()
    ioc_devices.ioc.start()
    yield ioc_devices
    # Purge the channel caches before we stop the IOC to stop
    # RuntimeError: Event loop is closed errors on teardown
    purge_channel_caches()
    ioc_devices.ioc.stop()
    # Print the IOC process output so in the case of a failing test
    # we will see if anything on the IOC side also failed
    print(ioc_devices.ioc.output)


class ExpectedData(Generic[T]):
    def __init__(
        self, initial: T, put: T, dtype: Dtype, dtype_numpy: str | list, **metadata
    ):
        self.initial = initial
        self.put = put
        self.metadata = dict(dtype=dtype, dtype_numpy=dtype_numpy, **metadata)


async def assert_monitor_then_put(
    signal: SignalR[SignalDatatypeT],
    initial_value: SignalDatatypeT,
    put_value: SignalDatatypeT,
    metadata: dict,
    signal_set: Callable[[SignalDatatypeT], Awaitable[None]] | None = None,
):
    if signal_set is None:
        assert isinstance(signal, SignalRW)
        signal_set = signal.set
    await signal.connect(timeout=1)
    with MonitorQueue(signal) as q:
        # Check initial value
        await q.assert_updates(initial_value)
        # Check descriptor
        if isinstance(initial_value, np.ndarray):
            shape = list(initial_value.shape)
        elif isinstance(initial_value, list | Table):
            shape = [len(initial_value)]
        else:
            shape = []
        await assert_describe_signal(signal, shape=shape, **metadata)
        # Put to new value and check it
        await signal_set(put_value)
        await q.assert_updates(put_value)


# Can be removed once numpy >=2 is pinned.
scalar_int_dtype = (
    "<i4" if os.name == "nt" and np.version.version.startswith("1.") else "<i8"
)
CA_PVA_INFERRED = {
    "a_int": ExpectedData(
        42,
        43,
        "integer",
        scalar_int_dtype,
        limits=Limits(
            control=LimitsRange(low=10, high=90),
            warning=LimitsRange(low=5, high=96),
            alarm=LimitsRange(low=2, high=98),
            display=LimitsRange(low=0, high=100),
        ),
        units="",
    ),
    "partialint": ExpectedData(
        42,
        43,
        "integer",
        scalar_int_dtype,
        limits=Limits(
            control=LimitsRange(low=10.0, high=90.0),
            alarm=LimitsRange(low=2.0, high=98.0),
            display=LimitsRange(low=0.0, high=100.0),
        ),
        units="",
    ),
    "lessint": ExpectedData(
        42,
        43,
        "integer",
        scalar_int_dtype,
        limits=Limits(
            # control = display if DRVL, DRVH not set
            control=LimitsRange(low=0.0, high=100.0),
            # LOPR, HOPR
            display=LimitsRange(low=0.0, high=100.0),
            # LOW, HIGH
            warning=LimitsRange(low=2.0, high=98.0),
        ),
        units="",
    ),
    "a_float": ExpectedData(3.141, 43.5, "number", "<f8", precision=1, units="mm"),
    "a_str": ExpectedData("hello", "goodbye", "string", "|S40"),
    "uint8a": ExpectedData(
        np.array([0, 255], dtype=np.uint8),
        np.array([218], dtype=np.uint8),
        "array",
        "|u1",
        units="",
    ),
    "int16a": ExpectedData(
        np.array([-32768, 32767], dtype=np.int16),
        np.array([-855], dtype=np.int16),
        "array",
        "<i2",
        units="",
    ),
    "int32a": ExpectedData(
        np.array([-2147483648, 2147483647], dtype=np.int32),
        np.array([-2], dtype=np.int32),
        "array",
        "<i4",
        units="",
    ),
    "float32a": ExpectedData(
        np.array([0.000002, -123.123], dtype=np.float32),
        np.array([1.0], dtype=np.float32),
        "array",
        "<f4",
        units="",
        precision=0,
    ),
    "float64a": ExpectedData(
        np.array([0.1, -12345678.123], dtype=np.float64),
        np.array([0.2], dtype=np.float64),
        "array",
        "<f8",
        units="",
        precision=0,
    ),
    "stra": ExpectedData(["five", "six", "seven"], ["nine", "ten"], "array", "|S40"),
}
PVA_INFERRED = {
    "int8a": ExpectedData(
        np.array([-128, 127, 0, 1, 2, 3, 4], dtype=np.int8),
        np.array([-8, 3, 44], dtype=np.int8),
        "array",
        "|i1",
        units="",
    ),
    "uint16a": ExpectedData(
        np.array([0, 65535, 0, 1, 2, 3, 4], dtype=np.uint16),
        np.array([5666], dtype=np.uint16),
        "array",
        "<u2",
        units="",
    ),
    "uint32a": ExpectedData(
        np.array([0, 4294967295, 0, 1, 2, 3, 4], dtype=np.uint32),
        np.array([1022233], dtype=np.uint32),
        "array",
        "<u4",
        units="",
    ),
    "int64a": ExpectedData(
        np.array([-(2**63 - 1), 2**63 - 1, 0, 1, 2, 3, 4], dtype=np.int64),
        np.array([-3], dtype=np.int64),
        "array",
        "<i8",
        units="",
    ),
    "uint64a": ExpectedData(
        np.array([0, 2**63 - 1, 0, 1, 2, 3, 4], dtype=np.uint64),
        np.array([995444], dtype=np.uint64),
        "array",
        "<u8",
        units="",
    ),
}


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize(
    "protocol,name,data",
    [("ca", k, v) for k, v in CA_PVA_INFERRED.items()]  # ca/pva shared for ca
    + [("pva", k, v) for k, v in CA_PVA_INFERRED.items()]  # ca/pva shared for pva
    + [("pva", k, v) for k, v in PVA_INFERRED.items()],  # pva only
)
async def test_epics_get_put_monitor_for_inferred_types(
    ioc_devices: EpicsTestIocAndDevices,
    protocol: Protocol,
    name: str,
    data: ExpectedData,
):
    # With the given datatype, check we have the correct initial value and putting
    # works
    device_signal = ioc_devices.get_signal(protocol, name)
    await assert_monitor_then_put(device_signal, data.initial, data.put, data.metadata)
    # With datatype guessed from CA/PVA, check we can set it back to the initial value
    guess_signal = epics_signal_rw(None, device_signal.source)  # type: ignore
    await assert_monitor_then_put(guess_signal, data.put, data.initial, data.metadata)


CA_PVA_OVERRIDE = {
    "longstr": ExpectedData(
        "a string that is just longer than forty characters",
        "another string that is just longer than forty characters",
        "string",
        "|S40",
    ),
    "longstr2": ExpectedData(
        "a string that is just longer than forty characters",
        "another string that is just longer than forty characters",
        "string",
        "|S40",
    ),
    "a_bool": ExpectedData(True, False, "boolean", dtype_numpy="|b1"),
    "bool_unnamed": ExpectedData(True, False, "boolean", dtype_numpy="|b1"),
    "enum": ExpectedData(
        EpicsTestEnum.B,
        EpicsTestEnum.C,
        "string",
        "|S40",
        choices=["Aaa", "Bbb", "Ccc"],
    ),
    "subset_enum": ExpectedData(
        EpicsTestSubsetEnum.B,
        EpicsTestSubsetEnum.A,
        "string",
        "|S40",
        choices=["Aaa", "Bbb", "Ccc"],
    ),
    "float_prec_0": ExpectedData(3, 4, "integer", scalar_int_dtype, units="mm"),
}
PVA_OVERRIDE = {}


@pytest.mark.timeout(TIMEOUT + 0.6)
@pytest.mark.parametrize(
    "protocol,name,data",
    [("ca", k, v) for k, v in CA_PVA_OVERRIDE.items()]  # ca/pva shared for ca
    + [("pva", k, v) for k, v in CA_PVA_OVERRIDE.items()]  # ca/pva shared for pva
    + [("pva", k, v) for k, v in PVA_OVERRIDE.items()],  # pva only
)
async def test_epics_get_put_monitor_for_override_types(
    ioc_devices: EpicsTestIocAndDevices,
    protocol: Protocol,
    name: str,
    data: ExpectedData,
):
    # With the given datatype, check we have the correct initial value and putting
    # works
    device_signal = ioc_devices.get_signal(protocol, name)
    await assert_monitor_then_put(device_signal, data.initial, data.put, data.metadata)
    # Then using the same signal, check that putting back works
    await assert_monitor_then_put(device_signal, data.put, data.initial, data.metadata)


def _example_table_dtype_numpy(guess: bool) -> list:
    return [
        # natively bool fields are uint8, so if we don't provide a Table
        # subclass to specify bool, that is what we get
        ("bool", "|u1" if guess else "|b1"),
        ("int", "<i4"),
        ("float", "<f8"),
        ("str", "|S40"),
        ("enum", "|S40"),
    ]


@pytest.mark.timeout(TIMEOUT)
async def test_pva_table(ioc_devices: EpicsTestIocAndDevices):
    initial = EpicsTestTable(
        a_bool=np.array([False, False, True, True], np.bool_),
        a_int=np.array([1, 8, -9, 32], np.int32),
        a_float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
        a_str=["Hello", "World", "Foo", "Bar"],
        a_enum=[EpicsTestEnum.A, EpicsTestEnum.B, EpicsTestEnum.A, EpicsTestEnum.C],
    )
    put = EpicsTestTable(
        a_bool=np.array([True, False], np.bool_),
        a_int=np.array([-5, 32], np.int32),
        a_float=np.array([8.5, -6.97], np.float64),
        a_str=["Hello", "Bat"],
        a_enum=[EpicsTestEnum.C, EpicsTestEnum.B],
    )
    dtype_numpy = [
        ("a_bool", "|b1"),
        ("a_int", "<i4"),
        ("a_float", "<f8"),
        ("a_str", "|S40"),
        ("a_enum", "|S40"),
    ]
    signal = ioc_devices.pva_device.table
    await assert_monitor_then_put(
        signal,
        initial,
        put,
        {"dtype": "array", "dtype_numpy": dtype_numpy},
    )
    initial_plain_table = Table(
        a_bool=np.array([0, 0, 1, 1], np.uint8),
        a_int=np.array([1, 8, -9, 32], np.int32),
        a_float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
        a_str=["Hello", "World", "Foo", "Bar"],
        a_enum=["Aaa", "Bbb", "Aaa", "Ccc"],
    )
    put_plain_table = Table(
        a_bool=np.array([1, 0], np.uint8),
        a_int=np.array([-5, 32], np.int32),
        a_float=np.array([8.5, -6.97], np.float64),
        a_str=["Hello", "Bat"],
        a_enum=["Ccc", "Bbb"],
    )
    dtype_numpy_plain_table = [
        # Plain tables will use the underlying epics datatype, in this
        # case uint8
        ("a_bool", "|u1"),
        ("a_int", "<i4"),
        ("a_float", "<f8"),
        ("a_str", "|S40"),
        ("a_enum", "|S40"),
    ]
    await assert_monitor_then_put(
        epics_signal_rw(None, signal.source),  # type: ignore
        put_plain_table,
        initial_plain_table,
        {"dtype": "array", "dtype_numpy": dtype_numpy_plain_table},
    )


@pytest.mark.timeout(TIMEOUT)
async def test_pva_ntndarray(ioc_devices: EpicsTestIocAndDevices):
    data = ExpectedData(np.zeros((2, 3)), np.arange(6).reshape((2, 3)), "array", "<i8")
    signal = ioc_devices.pva_device.ntndarray

    # Backdoor into the "raw" data underlying the NDArray in QSrv
    # not supporting direct writes to NDArray at the moment.
    raw_signal = epics_signal_rw(Array1D[np.int64], signal.source + ":data")
    await raw_signal.connect()

    async def signal_set(v):
        await raw_signal.set(v.flatten())

    await assert_monitor_then_put(
        signal, data.initial, data.put, data.metadata, signal_set
    )
    await assert_monitor_then_put(
        signal, data.put, data.initial, data.metadata, signal_set
    )


@pytest.mark.timeout(TIMEOUT)
async def test_writing_to_ndarray_raises_typeerror(ioc_devices: EpicsTestIocAndDevices):
    signal = epics_signal_rw(np.ndarray, ioc_devices.pva_device.ntndarray.source)
    await signal.connect()
    with pytest.raises(TypeError):
        await signal.set(np.zeros((6,), dtype=np.int64))


@pytest.mark.timeout(TIMEOUT)
async def test_invalid_enum_choice_raises_valueerror(
    ioc_devices: EpicsTestIocAndDevices,
):
    signal = ioc_devices.ca_device.enum_str_fallback
    await signal.connect()
    with pytest.raises(ValueError) as exc:
        await signal.set("Ddd")
    assert "Ddd is not a valid choice for" in str(exc.value)
    assert "ca:enum_str_fallback, valid choices: ['Aaa', 'Bbb', 'Ccc']" in str(
        exc.value
    )


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_typing_sequence_str_signal_connects(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    # Explicitly test that we can connect to a typing.Sequence[str] signal
    # rather than a collections.abc.Sequence[str] which is more normal
    signal = epics_signal_rw(typing.Sequence[str], ioc_devices.get_pv(protocol, "stra"))
    await signal.connect()


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_error_raised_on_disconnected_pv(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    signal = epics_signal_rw(bool, ioc_devices.get_pv(protocol, "bool"))
    await signal.connect()
    # The below will work without error
    await signal.set(False)
    # Override the PV so it fails
    signal._connector.backend.write_pv = "DisconnectedPv"  # type: ignore
    with pytest.raises(asyncio.TimeoutError, match=f"{protocol}://DisconnectedPv"):
        await signal.set(True, timeout=0.1)


class BadEnum(StrictEnum):
    A = "Aaa"
    B = "B"
    C = "Ccc"


class EnumNoString(Enum):
    A = "Aaa"


class SubsetEnumWrongChoices(SubsetEnum):
    A = "Aaa"
    B = "B"
    C = "Ccc"


def test_enum_equality():
    """Check that we are allowed to replace the passed datatype enum from a signal with
    a version generated from the signal with at least all of the same values, but
    possibly more.
    """

    class GeneratedChoices(StrictEnum):
        A = "Aaa"
        B = "B"
        C = "Ccc"

    class ExtendedGeneratedChoices(StrictEnum):
        A = "Aaa"
        B = "B"
        C = "Ccc"
        D = "Ddd"

    for enum_class in (GeneratedChoices, ExtendedGeneratedChoices):
        assert BadEnum.A == enum_class.A
        assert BadEnum.A.value == enum_class.A
        assert BadEnum.A.value == enum_class.A.value
        assert BadEnum(enum_class.A) is BadEnum.A
        assert BadEnum(enum_class.A.value) is BadEnum.A
        assert not BadEnum == enum_class

    # We will always PUT BadEnum by String, and GET GeneratedChoices by index,
    # so shouldn't ever run across this from conversion code, but may occur if
    # casting returned values or passing as enum rather than value.
    with pytest.raises(ValueError):
        BadEnum(ExtendedGeneratedChoices.D)


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
@pytest.mark.parametrize(
    "typ, suff, errors",
    [
        (
            BadEnum,
            "enum",
            (
                "has choices ('Aaa', 'Bbb', 'Ccc')",
                "but <enum 'BadEnum'>",
                "requested ['Aaa', 'B', 'Ccc'] to be strictly equal",
            ),
        ),
        (
            SubsetEnumWrongChoices,
            "enum",
            (
                "has choices ('Aaa', 'Bbb', 'Ccc')",
                "but <enum 'SubsetEnumWrongChoices'>",
                "requested ['Aaa', 'B', 'Ccc'] to be a subset",
            ),
        ),
        (
            int,
            "str",
            ("with inferred datatype str", "cannot be coerced to int"),
        ),
        (
            str,
            "float",
            ("with inferred datatype float", "cannot be coerced to str"),
        ),
        (
            str,
            "stra",
            ("with inferred datatype Sequence[str]", "cannot be coerced to str"),
        ),
        (
            int,
            "uint8a",
            ("with inferred datatype Array1D[np.uint8]", "cannot be coerced to int"),
        ),
        (
            float,
            "enum",
            ("with inferred datatype str", "cannot be coerced to float"),
        ),
        (
            Array1D[np.int32],
            "float64a",
            (
                "with inferred datatype Array1D[np.float64]",
                "cannot be coerced to Array1D[np.int32]",
            ),
        ),
        (
            EnumNoString,
            "enum2",
            (
                "<enum 'EnumNoString'> should inherit from ",
                "ophyd_async.core.SubsetEnum or ophyd_async.core.StrictEnum",
            ),
        ),
    ],
)
async def test_backend_wrong_type_errors(
    ioc_devices: EpicsTestIocAndDevices, typ, suff, errors, protocol: Protocol
):
    signal = epics_signal_rw(typ, ioc_devices.get_pv(protocol, suff))
    with pytest.raises(TypeError) as exc:
        await signal.connect()
    for error in errors:
        assert error in str(exc.value)


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_backend_put_enum_string(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    signal = ioc_devices.get_signal(protocol, "enum2")
    await signal.connect()
    await signal.set("Ccc")
    assert (
        Location(setpoint=EpicsTestEnum.C, readback=EpicsTestEnum.C)
        == await signal.locate()
    )
    val = await signal.get_value()
    assert val == "Ccc"
    assert val is EpicsTestEnum.C
    assert repr(val) == "<EpicsTestEnum.C: 'Ccc'>"


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_non_existent_errors(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    signal = epics_signal_rw(str, "non-existent")
    with pytest.raises(NotConnectedError):
        await signal.connect(timeout=0.1)


@pytest.mark.parametrize(
    "dt,expected",
    [
        (Array1D[np.int32], "Array1D[np.int32]"),
        (np.ndarray, "ndarray"),
        (npt.NDArray[np.float64], "Array1D[np.float64]"),
    ],
)
def test_format_error_message(dt, expected):
    assert format_datatype(dt) == expected


def test_make_backend_fails_for_different_transports():
    read_pv = "test"
    write_pv = "pva://test"

    with pytest.raises(
        TypeError,
        match=f"Differing protocols: {read_pv} has EpicsProtocol.CA,"
        + f" {write_pv} has EpicsProtocol.PVA",
    ):
        epics_signal_rw(str, read_pv, write_pv)


def _get_epics_backend(signal: Signal) -> CaSignalBackend | PvaSignalBackend:
    backend = signal._connector.backend
    assert isinstance(backend, CaSignalBackend | PvaSignalBackend)
    return backend


def test_signal_helpers():
    read_write = epics_signal_rw(int, "ReadWrite")
    assert _get_epics_backend(read_write).read_pv == "ReadWrite"
    assert _get_epics_backend(read_write).write_pv == "ReadWrite"

    read_write_rbv_manual = epics_signal_rw(int, "ReadWrite_RBV", "ReadWrite")
    assert _get_epics_backend(read_write_rbv_manual).read_pv == "ReadWrite_RBV"
    assert _get_epics_backend(read_write_rbv_manual).write_pv == "ReadWrite"

    read_write_rbv = epics_signal_rw_rbv(int, "ReadWrite")
    assert _get_epics_backend(read_write_rbv).read_pv == "ReadWrite_RBV"
    assert _get_epics_backend(read_write_rbv).write_pv == "ReadWrite"

    read_write_rbv_suffix = epics_signal_rw_rbv(int, "ReadWrite", read_suffix=":RBV")
    assert _get_epics_backend(read_write_rbv_suffix).read_pv == "ReadWrite:RBV"
    assert _get_epics_backend(read_write_rbv_suffix).write_pv == "ReadWrite"

    read_write_rbv_w_field = epics_signal_rw_rbv(int, "ReadWrite.VAL")
    assert _get_epics_backend(read_write_rbv_w_field).read_pv == "ReadWrite_RBV.VAL"
    assert _get_epics_backend(read_write_rbv_w_field).write_pv == "ReadWrite.VAL"

    read = epics_signal_r(int, "Read")
    assert _get_epics_backend(read).read_pv == "Read"

    write = epics_signal_w(int, "Write")
    assert _get_epics_backend(write).write_pv == "Write"

    execute = epics_signal_x("Execute")
    assert _get_epics_backend(execute).write_pv == "Execute"


def test_signal_helpers_explicit_read_timeout():
    # Check that we can adjust the _timeout attribute, which is used
    # for example during await signal.get_value()

    read_write = epics_signal_rw(int, "ReadWrite", timeout=123)
    assert read_write._timeout == 123

    read_write_rbv = epics_signal_rw_rbv(int, "ReadWrite", timeout=456)
    assert read_write_rbv._timeout == 456

    read = epics_signal_r(int, "Read", timeout=789)
    assert read._timeout == 789

    write = epics_signal_w(int, "Write", timeout=987)
    assert write._timeout == 987

    execute = epics_signal_x("Execute", timeout=654)
    assert execute._timeout == 654


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_signals_created_for_not_prec_0_float_cannot_use_int(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    sig = epics_signal_rw(int, ioc_devices.get_pv(protocol, "float_prec_1"))
    with pytest.raises(
        TypeError,
        match="float_prec_1 with inferred datatype float.* cannot be coerced to int",
    ):
        await sig.connect()


@pytest.mark.timeout(TIMEOUT)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_bool_works_for_mismatching_enums(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    pv_name = ioc_devices.get_pv(protocol, "bool")
    sig = epics_signal_rw(bool, pv_name, pv_name + "_unnamed")
    await sig.connect()


@pytest.mark.timeout(TIMEOUT)
async def test_can_read_using_ophyd_async_then_ophyd(
    RE, ioc_devices: EpicsTestIocAndDevices
):
    ophyd_async_sig = epics_signal_rw(float, ioc_devices.get_pv("ca", "float_prec_1"))
    await ophyd_async_sig.connect()
    ophyd_signal = EpicsSignal(ioc_devices.get_pv("ca", "float_prec_0").split("://")[1])
    ophyd_signal.wait_for_connection(timeout=5)

    def a_plan():
        yield from bps.rd(ophyd_async_sig)
        yield from bps.rd(ophyd_signal)

    RE(a_plan())


def test_signal_module_emits_deprecation_warning():
    with pytest.deprecated_call():
        import ophyd_async.epics.signal  # noqa: F401


@pytest.mark.timeout(TIMEOUT + 0.6)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_observe_ticking_signal_with_busy_loop(
    ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    sig = epics_signal_rw(int, ioc_devices.get_pv("ca", "ticking"))
    await sig.connect()

    recv = []

    async def watch():
        async for val in observe_value(sig, done_timeout=0.4):
            time.sleep(0.3)
            recv.append(val)

    start = time.monotonic()

    with pytest.raises(asyncio.TimeoutError):
        await watch()
    assert time.monotonic() - start == pytest.approx(0.6, abs=0.1)
    assert len(recv) == 2
    # Don't check values as CA and PVA have different algorithms for
    # dropping updates for slow callbacks


@pytest.mark.parametrize(
    "all_updates,expected_all_updates",
    [("True", True), ("False", False)],
)
@pytest.mark.timeout(TIMEOUT + 0.6)
async def test_all_updates_settable_with_environment_variable(
    ioc_devices: EpicsTestIocAndDevices,
    all_updates: Any,
    expected_all_updates: bool,
):
    with patch.dict(
        os.environ,
        {**os.environ, "OPHYD_ASYNC_EPICS_CA_KEEP_ALL_UPDATES": all_updates},
        clear=True,
    ):
        assert (
            await default_all_updates(ioc_devices.get_pv("ca", "ticking"))
        ) is expected_all_updates


@pytest.mark.timeout(TIMEOUT + 0.6)
async def test_all_updates_defaults_to_true(
    ioc_devices: EpicsTestIocAndDevices,
):
    assert await default_all_updates(ioc_devices.get_pv("ca", "ticking"))


async def default_all_updates(pv: str) -> bool:
    sig = epics_signal_rw(int, pv)
    await sig.connect()
    backend = sig._connector.backend

    try:
        sig.subscribe_reading(lambda v: ...)
        assert isinstance(backend.subscription, Subscription)
        assert isinstance(backend.subscription.all_updates, bool)
        return backend.subscription.all_updates
    finally:
        backend.subscription.close()
        sig.clear_sub(lambda v: ...)


HERE = Path(__file__).absolute().parent


@pytest.mark.timeout(TIMEOUT + 0.5)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_retrieve_apply_store_settings(
    RE, ioc_devices: EpicsTestIocAndDevices, protocol: Protocol, tmp_path
):
    tmp_provider = YamlSettingsProvider(tmp_path)
    expected_provider = YamlSettingsProvider(HERE)
    device = ioc_devices.get_device(protocol)

    def a_plan():
        yield from ensure_connected(device)
        settings = yield from retrieve_settings(
            expected_provider, f"test_yaml_save_{protocol}", device
        )
        yield from apply_settings(settings)
        yield from store_settings(tmp_provider, "test_file", device)
        with open(tmp_path / "test_file.yaml") as actual_file:
            with open(HERE / f"test_yaml_save_{protocol}.yaml") as expected_file:
                # If this test fails because you added a signal, then you can regenerate
                # the test data with:
                # cp /tmp/pytest-of-root/pytest-current/test_retrieve_apply_store_sett0/test_file.yaml tests/epics/signal/test_yaml_save_ca.yaml  # noqa: E501
                # cp /tmp/pytest-of-root/pytest-current/test_retrieve_apply_store_sett1/test_file.yaml tests/epics/signal/test_yaml_save_pva.yaml  # noqa: E501
                assert yaml.safe_load(actual_file) == yaml.safe_load(expected_file)

    RE(a_plan())


@pytest.mark.timeout(TIMEOUT + 0.5)
@pytest.mark.parametrize("protocol", get_args(Protocol))
async def test_put_completion(
    RE, ioc_devices: EpicsTestIocAndDevices, protocol: Protocol
):
    # Check that we can put to an epics signal and wait for put completion
    slow_seq_pv = ioc_devices.get_pv(protocol, "slowseq")
    slow_seq = epics_signal_rw(int, slow_seq_pv, wait=lambda value: value == 1)
    await slow_seq.connect()

    # First, do a set with blocking and make sure it takes a while
    start = time.monotonic()
    await slow_seq.set(1)
    stop = time.monotonic()
    assert stop - start == pytest.approx(0.5, rel=0.1)

    # Then, make sure if we don't wait it returns ~instantly
    start = time.monotonic()
    await slow_seq.set(2)
    stop = time.monotonic()
    assert stop - start < 0.1

    # Time for completion callback to have finished before moving to
    # next test / iteration - without this, running this test multiple
    # times in a row will fail even-numbered runs.
    await asyncio.sleep(0.5)


@pytest.mark.timeout(TIMEOUT + 0.5)
async def test_setting_with_none_uses_initial_value_of_pv(
    ioc_devices: EpicsTestIocAndDevices,
):
    sig_rw = epics_signal_rw(int, ioc_devices.get_pv("pva", "slowseq"))
    await sig_rw.connect()
    initial_data = await sig_rw.read()
    initial_value, initial_timestamp = (
        initial_data[""]["value"],
        initial_data[""]["timestamp"],
    )

    # This mimics triggering a SignalX
    await sig_rw.set(None)  # type: ignore

    current_data = await sig_rw.read()
    assert (
        initial_value == current_data[""]["value"]
        and initial_timestamp != current_data[""]["timestamp"]
    )


@pytest.mark.timeout(TIMEOUT + 0.5)
async def test_signal_retries_when_timeout(
    ioc_devices: EpicsTestIocAndDevices,
):
    # put callback on slowseq in 0.5s, so if waited, this will fail to set
    sig_rw_times_out = epics_signal_rw(
        int, ioc_devices.get_pv("pva", "slowseq"), attempts=3, timeout=0.1
    )
    await sig_rw_times_out.connect()

    start = time.monotonic()
    with pytest.raises(asyncio.TimeoutError):
        await sig_rw_times_out.set(1)
    stop = time.monotonic()
    # signal tries to set 3 times, so 3 * timeout
    assert stop - start == pytest.approx(0.3, abs=0.1)


async def test_signal_timestamp_is_same_format_as_soft_signal_timestamp(
    RE, ioc_devices: EpicsTestIocAndDevices
):
    sim_sig, sim_sig_setter = soft_signal_r_and_setter(float)
    real_sig = epics_signal_rw(float, ioc_devices.get_pv("ca", "float_prec_1"))
    await real_sig.connect(timeout=30)

    await real_sig.set(10)
    sim_sig_setter(20)

    real_data = await real_sig.read()
    sim_data = await sim_sig.read()

    assert abs(real_data[""]["timestamp"] - sim_data[""]["timestamp"]) < 0.1


@pytest.mark.parametrize("protocol", get_args(Protocol))
@pytest.mark.parametrize("mock", [True, False])
def test_subscribe_works_under_re_and_fails_outside(
    RE, ioc_devices: EpicsTestIocAndDevices, protocol: Protocol, mock: bool
):
    s1 = epics_signal_r(float, ioc_devices.get_pv(protocol, "float"), "s1")
    s2 = epics_signal_r(float, ioc_devices.get_pv(protocol, "float"), "s2")

    # Run in the RE event loop
    def plan():
        yield from ensure_connected(s1, s2, mock=mock)
        if mock:
            set_mock_value(s1, 3.141)
        q = asyncio.Queue()
        s1.subscribe_reading(q.put_nowait)
        try:
            (fut,) = yield from bps.wait_for([q.get])
            assert fut.result()["s1"]["value"] == 3.141
        finally:
            s1.clear_sub(q.put_nowait)

    RE(plan())

    # Run plan outside RE and outside event loop
    with pytest.raises(
        RuntimeError,
        match="Need a running event loop to subscribe to a signal, "
        "are you trying to run subscribe outside a plan?",
    ):
        s2.subscribe_reading(print)



import asyncio
import re
import sys
from asyncio import CancelledError

import pytest

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    gather_dict,
    init_devices,
    SoftSignalBackend,
    NotConnectedError,
    SignalRW,
    AsyncStatus,
    get_mock_put,
    set_mock_value,
    soft_signal_rw,
)
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.epics.motor import Motor


class ValueErrorBackend(SoftSignalBackend):
    def __init__(self, exc_text=""):
        self.exc_text = exc_text
        super().__init__(datatype=int)

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        raise ValueError(self.exc_text)


class WorkingDummyChildDevice(Device):
    def __init__(self, name: str = "working_dummy_child_device") -> None:
        self.working_signal = soft_signal_rw(int)
        super().__init__(name=name)


class TimeoutDummyChildDeviceCA(Device):
    def __init__(self, name: str = "timeout_dummy_child_device_ca") -> None:
        self.timeout_signal = epics_signal_rw(int, "ca://A_NON_EXISTENT_SIGNAL")
        super().__init__(name=name)


class TimeoutDummyChildDevicePVA(Device):
    def __init__(self, name: str = "timeout_dummy_child_device_pva") -> None:
        self.timeout_signal = epics_signal_rw(int, "pva://A_NON_EXISTENT_SIGNAL")
        super().__init__(name=name)


class ValueErrorDummyChildDevice(Device):
    def __init__(
        self, name: str = "value_error_dummy_child_device", exc_text=""
    ) -> None:
        self.value_error_signal = SignalRW(ValueErrorBackend(exc_text=exc_text))
        super().__init__(name=name)


class DummyDeviceOneWorkingOneTimeout(Device):
    def __init__(self, name: str = "dummy_device_one_working_one_timeout") -> None:
        self.working_child_device = WorkingDummyChildDevice()
        self.timeout_child_device = TimeoutDummyChildDeviceCA()
        super().__init__(name=name)


ONE_WORKING_ONE_TIMEOUT_OUTPUT = NotConnectedError(
    {
        "timeout_child_device": NotConnectedError(
            {"timeout_signal": NotConnectedError("ca://A_NON_EXISTENT_SIGNAL")}
        )
    }
)


class DummyDeviceTwoWorkingTwoTimeOutTwoValueError(Device):
    def __init__(
        self,
        name: str = "dummy_device_two_working_one_timeout_two_value_error",
    ) -> None:
        self.working_child_device1 = WorkingDummyChildDevice()
        self.working_child_device2 = WorkingDummyChildDevice()
        self.timeout_child_device_ca = TimeoutDummyChildDeviceCA()
        self.timeout_child_device_pva = TimeoutDummyChildDevicePVA()
        self.value_error_child_device1 = ValueErrorDummyChildDevice(
            exc_text="Some ValueError text"
        )
        self.value_error_child_device2 = ValueErrorDummyChildDevice()
        super().__init__(name=name)


TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT = NotConnectedError(
    {
        "timeout_child_device_ca": NotConnectedError(
            {
                "timeout_signal": NotConnectedError("ca://A_NON_EXISTENT_SIGNAL"),
            }
        ),
        "timeout_child_device_pva": NotConnectedError(
            {"timeout_signal": NotConnectedError("pva://A_NON_EXISTENT_SIGNAL")}
        ),
        "value_error_child_device1": NotConnectedError(
            {"value_error_signal": ValueError("Some ValueError text")}
        ),
        "value_error_child_device2": NotConnectedError(
            {
                "value_error_signal": ValueError(),
            }
        ),
    }
)


class DummyDeviceCombiningTopLevelSignalAndSubDevice(Device):
    def __init__(
        self, name: str = "dummy_device_combining_top_level_signal_and_sub_device"
    ) -> None:
        self.timeout_signal = epics_signal_rw(int, "ca://A_NON_EXISTENT_SIGNAL")
        self.sub_device = ValueErrorDummyChildDevice(exc_text="Some ValueError text")
        super().__init__(name=name)


async def test_error_handling_connection_timeout(caplog):
    caplog.set_level(10)

    dummy_device_one_working_one_timeout = DummyDeviceOneWorkingOneTimeout()

    # This should work since the error is a connection timeout
    with pytest.raises(NotConnectedError) as exc:
        await dummy_device_one_working_one_timeout.connect(timeout=0.01)

    assert str(exc.value) == str(ONE_WORKING_ONE_TIMEOUT_OUTPUT)

    logs = caplog.get_records("call")

    # See https://github.com/bluesky/ophyd-async/issues/519
    # assert len(logs) == 3

    assert "signal ca://A_NON_EXISTENT_SIGNAL timed out" == logs[-1].message
    assert logs[-1].levelname == "DEBUG"


async def test_error_handling_value_errors(caplog):
    """Checks that NotConnectedError is aggregated correctly across Devices."""

    caplog.set_level(10)

    dummy_device_two_working_one_timeout_two_value_error = (
        DummyDeviceTwoWorkingTwoTimeOutTwoValueError("dsf")
    )

    # This should fail since the error is a ValueError
    with pytest.raises(NotConnectedError) as exc:
        await dummy_device_two_working_one_timeout_two_value_error.connect(timeout=0.01)

    assert str(exc.value) == str(TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT)

    logs = caplog.get_records("call")
    logs = [
        log
        for log in logs
        if "ophyd_async" in log.pathname and "_signal" not in log.pathname
    ]
    assert len(logs) == 4

    for i in range(0, 2):
        assert (
            "device `value_error_signal` raised unexpected exception ValueError"
            == logs[i].message
        )
        assert logs[i].levelname == "ERROR"

    assert logs[2].levelname == "DEBUG"
    assert logs[3].levelname == "DEBUG"
    # These messages could come in any order
    messages = [logs[idx].message for idx in (2, 3)]
    for protocol in ("pva", "ca"):
        assert f"signal {protocol}://A_NON_EXISTENT_SIGNAL timed out" in messages


class BadDatatypeDevice(Device):
    def __init__(self, name: str = "") -> None:
        self.sig = epics_signal_rw(object, "", "")
        super().__init__(name)


async def test_error_handling_init_devices_mock():
    with pytest.raises(NotConnectedError) as exc:
        async with init_devices(mock=True):
            device = BadDatatypeDevice()
            device2 = BadDatatypeDevice()
    expected_output = NotConnectedError(
        {
            "device": NotConnectedError(
                {"sig": TypeError(f"Can't make converter for {object}")}
            ),
            "device2": NotConnectedError(
                {"sig": TypeError(f"Can't make converter for {object}")}
            ),
        }
    )
    assert str(expected_output) == str(exc.value)


def test_introspecting_sub_errors():
    sub_error1 = NotConnectedError("bad")
    assert sub_error1.sub_errors == {}
    sub_error2 = ValueError("very bad")
    error = NotConnectedError({"child1": sub_error1, "child2": sub_error2})
    assert error.sub_errors == {"child1": sub_error1, "child2": sub_error2}


async def test_error_handling_init_devices(caplog):
    caplog.set_level(10)
    with pytest.raises(NotConnectedError) as exc:
        # flake8: noqa
        async with init_devices(timeout=0.1):
            dummy_device_two_working_one_timeout_two_value_error = (
                DummyDeviceTwoWorkingTwoTimeOutTwoValueError()
            )
            dummy_device_one_working_one_timeout = DummyDeviceOneWorkingOneTimeout()

    expected_output = NotConnectedError(
        {
            "dummy_device_two_working_one_timeout_two_value_error": (
                TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT
            ),
            "dummy_device_one_working_one_timeout": ONE_WORKING_ONE_TIMEOUT_OUTPUT,
        }
    )
    assert str(expected_output) == str(exc.value)

    logs = caplog.get_records("call")
    logs = [
        log
        for log in logs
        if "ophyd_async" in log.pathname and "_signal" not in log.pathname
    ]
    assert len(logs) == 5
    assert (
        logs[0].message
        == logs[1].message
        == "device `value_error_signal` raised unexpected exception ValueError"
    )
    assert logs[0].levelname == logs[1].levelname == "ERROR"

    assert logs[2].levelname == "DEBUG"
    assert logs[3].levelname == "DEBUG"
    # These messages could come in any order
    messages = [logs[idx].message for idx in (2, 3)]
    for protocol in ("pva", "ca"):
        assert f"signal {protocol}://A_NON_EXISTENT_SIGNAL timed out" in messages


def test_not_connected_error_output():
    assert str(TWO_WORKING_TWO_TIMEOUT_TWO_VALUE_ERROR_OUTPUT) == (
        "\ntimeout_child_device_ca: NotConnectedError:\n"
        "    timeout_signal: NotConnectedError: ca://A_NON_EXISTENT_SIGNAL\n"
        "timeout_child_device_pva: NotConnectedError:\n"
        "    timeout_signal: NotConnectedError: pva://A_NON_EXISTENT_SIGNAL\n"
        "value_error_child_device1: NotConnectedError:\n"
        "    value_error_signal: ValueError: Some ValueError text\n"
        "value_error_child_device2: NotConnectedError:\n"
        "    value_error_signal: ValueError\n"
    )


async def test_combining_top_level_signal_and_child_device():
    dummy_device1 = DummyDeviceCombiningTopLevelSignalAndSubDevice()
    with pytest.raises(NotConnectedError) as exc:
        await dummy_device1.connect(timeout=0.01)
    assert str(exc.value) == (
        "\ntimeout_signal: NotConnectedError: ca://A_NON_EXISTENT_SIGNAL\n"
        "sub_device: NotConnectedError:\n"
        "    value_error_signal: ValueError: Some ValueError text\n"
    )

    with pytest.raises(NotConnectedError) as exc:
        async with init_devices(timeout=0.1):
            dummy_device2 = DummyDeviceCombiningTopLevelSignalAndSubDevice()
    assert str(exc.value) == (
        "\ndummy_device2: NotConnectedError:\n"
        "    timeout_signal: NotConnectedError: ca://A_NON_EXISTENT_SIGNAL\n"
        "    sub_device: NotConnectedError:\n"
        "        value_error_signal: ValueError: Some ValueError text\n"
    )


async def test_format_error_string_input():
    with pytest.raises(
        RuntimeError,
        match=("Unexpected type `<class 'int'>` expected `str` or `dict`"),
    ):
        not_connected = NotConnectedError(123)
        str(not_connected)

    with pytest.raises(
        RuntimeError, match=("Unexpected type `<class 'int'>`, expected an Exception")
    ):
        not_connected = NotConnectedError({"test": 123})
        str(not_connected)


def test_core_notconnected_emits_deprecation_warning():
    with pytest.deprecated_call():
        from ophyd_async.core import NotConnected  # noqa: F401


async def _async_value(val):
    """Helper function to create an awaitable that returns a value."""
    return val


async def test_gather_dict_empty():
    """Test gather_dict with an empty dictionary."""
    result = await gather_dict({})
    assert result == {}


async def test_gather_dict_duplicate_resolved_keys():
    """Test gather_dict when different awaitable keys resolve to the same value."""
    # Two coroutine objects that resolve to the same key - second should win
    result = await gather_dict(
        {
            _async_value("key"): "first",
            _async_value("key"): "second",
        }
    )
    assert result == {"key": "second"}


@pytest.mark.parametrize("keys_awaitable", ("", "b", "abc"))
@pytest.mark.parametrize("values_awaitable", ([], [2], [1, 2, 3]))
async def test_gather_dict(keys_awaitable, values_awaitable):
    """Test gather_dict with plain values, awaitable keys, and awaitable values."""
    expected = {"a": 1, "b": 2, "c": 3}

    input_dict = {}
    for k, v in expected.items():
        key = _async_value(k) if k in keys_awaitable else k
        value = _async_value(v) if v in values_awaitable else v
        input_dict[key] = value
    result = await gather_dict(input_dict)
    assert result == expected
    assert list(result.keys()) == list(expected.keys())


# Cancellation propagation is broken in asyncio.gather on Python 3.10, so the exception
# with the message is not reachable via the __cause__ chain however you will still get
# the message in the stack trace
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="https://github.com/python/cpython/issues/112534"
)
@pytest.mark.parametrize(
    "set_to_delay, expected_message, expected_results",
    # Test cases for different things taking too long
    [
        # plain awaitable (currently not handled)
        [
            "x",
            None,
            None,
        ],
        # Device with a custom set() method
        ["y", "CancelledError while awaiting .* on my_device", None],
        # Task wrapping an async function (currently not handled)
        ["z", None, None],
        # Setting a Motor on a Device
        ["omega", "BL03I-MO-OMEGA", None],
        # Everything completes successfully in a timely fashion
        [None, None, [1, None, 1, None]],
    ],
)
async def test_cancelled_error_message_for_gather_is_populated_on_timeout(
    set_to_delay: str, expected_message: str, expected_results: list
):
    async def async_func(name: str, value: int) -> int:
        if set_to_delay == name:
            await asyncio.sleep(20)
        return value

    plain_awaitable = async_func("x", 1)

    class GenericDevice(Device):
        def __init__(self, name):
            super().__init__(name)

        @AsyncStatus.wrap
        async def set(self, value: int):
            await async_func("y", value)

    class MotorDevice(Device):
        def __init__(self, name):
            self.omega = Motor("BL03I-MO-OMEGA")
            super().__init__(name)

        async def set_omega(self, value: int, *args, **kwargs):
            await async_func("omega", value)

    async_status_device = GenericDevice("my_device")
    task_wrapped = asyncio.create_task(async_func("z", 1), name="set z")
    motor_device = MotorDevice("my_motor_device")
    await motor_device.connect(mock=True)
    _patch_motor(motor_device.omega)
    get_mock_put(motor_device.omega.user_setpoint).side_effect = motor_device.set_omega

    gather_awaitable = asyncio.gather(
        plain_awaitable,
        async_status_device.set(1),
        task_wrapped,
        motor_device.omega.set(1),
    )
    if set_to_delay:
        with pytest.raises(asyncio.TimeoutError) as exc_info:
            await asyncio.wait_for(gather_awaitable, timeout=0.3)
        cause = exc_info.value.__cause__
        assert isinstance(cause, CancelledError)
        assert _contains_message(cause, expected_message)
    else:
        results = await gather_awaitable
        assert results == expected_results


def _contains_message(e: BaseException, expected_message: str) -> bool:
    return (
        expected_message is None
        or (e.args and re.search(expected_message, e.args[0]))
        or (e.__cause__ and _contains_message(e.__cause__, expected_message))
    )


def _patch_motor(motor: Motor, initial_position=0):
    set_mock_value(motor.user_setpoint, initial_position)
    set_mock_value(motor.user_readback, initial_position)
    set_mock_value(motor.deadband, 0.001)
    set_mock_value(motor.motor_done_move, 1)
    set_mock_value(motor.velocity, 3)
    set_mock_value(motor.max_velocity, 5)
    set_mock_value(motor.high_limit_travel, float("inf"))
    set_mock_value(motor.low_limit_travel, float("-inf"))



from __future__ import annotations

import asyncio
import atexit
import logging
import typing
from collections.abc import Mapping, Sequence
from math import isnan, nan
from typing import Any, Generic

import numpy as np
from bluesky.protocols import Reading
from event_model import DataKey, Limits, LimitsRange
from p4p import Value
from p4p.client.asyncio import Context, Subscription
from pydantic import BaseModel

from ophyd_async.core import (
    Array1D,
    Callback,
    NotConnectedError,
    SignalDatatype,
    SignalDatatypeT,
    SignalMetadata,
    Table,
    get_enum_cls,
    get_unique,
    make_datakey,
    wait_for_connection,
)

from ._util import (
    EpicsOptions,
    EpicsSignalBackend,
    format_datatype,
    get_supported_values,
)

logger = logging.getLogger("ophyd_async")


def _limits_from_value(value: Any) -> Limits:
    def get_limits(
        substucture_name: str, low_name: str = "limitLow", high_name: str = "limitHigh"
    ) -> LimitsRange | None:
        substructure = getattr(value, substucture_name, None)
        low = getattr(substructure, low_name, nan)
        high = getattr(substructure, high_name, nan)
        if not (isnan(low) and isnan(high)) and not low == high == 0:
            return LimitsRange(
                low=None if isnan(low) else low,
                high=None if isnan(high) else high,
            )
        return None

    limits = Limits()
    if limits_range := get_limits("valueAlarm", "lowAlarmLimit", "highAlarmLimit"):
        limits["alarm"] = limits_range
    if limits_range := get_limits("control"):
        limits["control"] = limits_range
    if limits_range := get_limits("display"):
        limits["display"] = limits_range
    if limits_range := get_limits("valueAlarm", "lowWarningLimit", "highWarningLimit"):
        limits["warning"] = limits_range
    return limits


def _metadata_from_value(datatype: type[SignalDatatype], value: Any) -> SignalMetadata:
    metadata = SignalMetadata()
    value_data: Any = getattr(value, "value", None)
    specifier = _get_specifier(value)
    display_data: Any = getattr(value, "display", None)
    if (
        hasattr(display_data, "units")
        and specifier[-1] in _number_specifiers
        and datatype is not str
    ):
        metadata["units"] = display_data.units
    if (
        hasattr(display_data, "precision")
        and not isnan(display_data.precision)
        and specifier[-1] in _float_specifiers
        and datatype is not int
    ):
        metadata["precision"] = display_data.precision
    if (limits := _limits_from_value(value)) and specifier[-1] in _number_specifiers:
        metadata["limits"] = limits
    # Get choices from display or value
    if datatype is str or get_enum_cls(datatype) is not None:
        if hasattr(display_data, "choices"):
            metadata["choices"] = display_data.choices
        elif hasattr(value_data, "choices"):
            metadata["choices"] = value_data.choices
    return metadata


class PvaConverter(Generic[SignalDatatypeT]):
    value_fields = ("value",)
    reading_fields = ("alarm", "timeStamp")

    def __init__(self, datatype: type[SignalDatatypeT]):
        self.datatype = datatype

    def value(self, value: Any) -> SignalDatatypeT:
        # Normally the value will be of the correct python type
        return value["value"]

    def write_value(self, value: Any) -> Any:
        # The pva library will do the conversion for us
        return value


class PvaIntConverter(PvaConverter[int]):
    def __init__(self):
        super().__init__(int)

    def value(self, value: Any) -> int:
        # Convert to an int
        return int(value["value"])


class PvaLongStringConverter(PvaConverter[str]):
    def __init__(self):
        super().__init__(str)

    def value(self, value: Any) -> Any:
        # Value here is a null terminated array of ascii codes.
        # We strip out the null terminator, and convert each code
        # to the corresponding char, joining into a string
        return value["value"].tobytes().rstrip(b"\0").decode()

    def write_value(self, value: Any) -> Any:
        # Inverse of reading - convert each character into it's ascii code,
        # put into a list, and add null terminator.
        return np.frombuffer(str(value).encode() + b"\0", dtype=np.int8)


class DisconnectedPvaConverter(PvaConverter):
    def __getattribute__(self, __name: str) -> Any:
        raise NotImplementedError("No PV has been set as connect() has not been called")


class PvaNDArrayConverter(PvaConverter[SignalDatatypeT]):
    value_fields = ("value", "dimension")

    def _get_dimensions(self, value) -> list[int]:
        dimensions: list[Value] = value["dimension"]
        dims = [dim.size for dim in dimensions]
        # Note: dimensions in NTNDArray are in fortran-like order
        # with first index changing fastest.
        #
        # Therefore we need to reverse the order of the dimensions
        # here to get back to a more usual C-like order with the
        # last index changing fastest.
        return dims[::-1]

    def value(self, value: Any) -> SignalDatatypeT:
        dims = self._get_dimensions(value)
        return value["value"].reshape(dims)

    def write_value(self, value: Any) -> Any:
        # No clear use-case for writing directly to an NDArray, and some
        # complexities around flattening to 1-D - e.g. dimension-order.
        # Don't support this for now.
        raise TypeError("Writing to NDArray not supported")


class PvaEnumConverter(PvaConverter[str]):
    def __init__(
        self, datatype: type[str] = str, supported_values: Mapping[str, str] = {}
    ):
        self.supported_values = supported_values
        super().__init__(datatype)

    def value(self, value: Any) -> str:
        str_value = value["value"]["choices"][value["value"]["index"]]
        if self.supported_values:
            return self.supported_values[str_value]
        else:
            return str_value


class PvaEnumBoolConverter(PvaConverter[bool]):
    def __init__(self):
        super().__init__(bool)

    def value(self, value: Any) -> bool:
        return bool(value["value"]["index"])


class PvaTableConverter(PvaConverter[Table]):
    def value(self, value) -> Table:
        return self.datatype(**value["value"].todict())

    def write_value(self, value: BaseModel | dict[str, Any]) -> Any:
        if isinstance(value, self.datatype):
            return value.model_dump(mode="python")
        return value


# https://mdavidsaver.github.io/p4p/values.html
_float_specifiers = {"f", "d"}
_int_specifiers = {"b", "B", "h", "H", "i", "I", "l", "L"}
_number_specifiers = _float_specifiers.union(_int_specifiers)
_datatype_converter_from_typeid: dict[
    tuple[str, str], tuple[type[SignalDatatype], type[PvaConverter]]
] = {
    ("epics:nt/NTScalar:1.0", "?"): (bool, PvaConverter),
    ("epics:nt/NTScalar:1.0", "b"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "B"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "h"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "H"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "i"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "I"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "l"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "L"): (int, PvaConverter),
    ("epics:nt/NTScalar:1.0", "f"): (float, PvaConverter),
    ("epics:nt/NTScalar:1.0", "d"): (float, PvaConverter),
    ("epics:nt/NTScalar:1.0", "s"): (str, PvaConverter),
    ("epics:nt/NTEnum:1.0", "S"): (str, PvaEnumConverter),
    ("epics:nt/NTScalarArray:1.0", "a?"): (Array1D[np.bool_], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "ab"): (Array1D[np.int8], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "aB"): (Array1D[np.uint8], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "ah"): (Array1D[np.int16], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "aH"): (Array1D[np.uint16], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "ai"): (Array1D[np.int32], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "aI"): (Array1D[np.uint32], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "al"): (Array1D[np.int64], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "aL"): (Array1D[np.uint64], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "af"): (Array1D[np.float32], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "ad"): (Array1D[np.float64], PvaConverter),
    ("epics:nt/NTScalarArray:1.0", "as"): (Sequence[str], PvaConverter),
    ("epics:nt/NTTable:1.0", "S"): (Table, PvaTableConverter),
    ("epics:nt/NTNDArray:1.0", "v"): (np.ndarray, PvaNDArrayConverter),
    ("epics:nt/NTNDArray:1.0", "U"): (np.ndarray, PvaNDArrayConverter),
}


def _get_specifier(value: Value) -> str:
    typ = value.type("value").aspy()
    if isinstance(typ, tuple):
        return typ[0]
    else:
        return str(typ)


def make_converter(datatype: type | None, values: dict[str, Any]) -> PvaConverter:
    pv = list(values)[0]
    typeid = get_unique({k: v.getID() for k, v in values.items()}, "typeids")
    specifier = get_unique(
        {k: _get_specifier(v) for k, v in values.items()},
        "value type specifiers",
    )
    # Make the datatype canonical for the inference below
    if datatype == typing.Sequence[str]:
        datatype = Sequence[str]
    # Infer a datatype and converter from the typeid and specifier
    inferred_datatype, converter_cls = _datatype_converter_from_typeid[
        (typeid, specifier)
    ]
    # Some override cases
    if datatype is bool and typeid == "epics:nt/NTEnum:1.0":
        # Database can't do bools, so are often representated as enums of len 2
        pv_num_choices = get_unique(
            {k: len(v["value"]["choices"]) for k, v in values.items()},
            "number of choices",
        )
        if pv_num_choices != 2:
            raise TypeError(f"{pv} has {pv_num_choices} choices, can't map to bool")
        return PvaEnumBoolConverter()
    elif typeid == "epics:nt/NTEnum:1.0":
        pv_choices = get_unique(
            {k: tuple(v["value"]["choices"]) for k, v in values.items()}, "choices"
        )
        if enum_cls := get_enum_cls(datatype):
            # We were given an enum class, so make class from that
            return PvaEnumConverter(
                supported_values=get_supported_values(pv, enum_cls, pv_choices)
            )
        elif datatype in (None, str):
            # Still use the Enum converter, but make choices from what it has
            return PvaEnumConverter()
    elif (
        inferred_datatype is float
        and datatype is int
        and get_unique(
            {k: v["display"]["precision"] for k, v in values.items()}, "precision"
        )
        == 0
    ):
        # Allow int signals to represent float records when prec is 0
        return PvaIntConverter()
    elif inferred_datatype is str and (enum_cls := get_enum_cls(datatype)):
        # Allow strings to be used as enums until QSRV supports this
        return PvaConverter(str)
    elif inferred_datatype is Table and datatype and issubclass(datatype, Table):
        # Use a custom table class
        return PvaTableConverter(datatype)
    elif datatype in (None, inferred_datatype):
        # If datatype matches what we are given then allow it and use inferred converter
        return converter_cls(inferred_datatype)
    # Allow waveforms with FTVL=CHAR to be treated as str when requested
    elif datatype is str and inferred_datatype == Array1D[np.int8]:
        return PvaLongStringConverter()
    raise TypeError(
        f"{pv} with inferred datatype {format_datatype(inferred_datatype)}"
        f" from {typeid=} {specifier=}"
        f" cannot be coerced to {format_datatype(datatype)}"
    )


_context: Context | None = None


def context() -> Context:
    global _context
    if _context is None:
        _context = Context("pva", nt=False)

        @atexit.register
        def _del_ctxt():
            # If we don't do this we get messages like this on close:
            #   Error in sys.excepthook:
            #   Original exception was:
            global _context
            del _context

    return _context


async def pvget_with_timeout(pv: str, timeout: float) -> Any:
    try:
        return await asyncio.wait_for(context().get(pv), timeout=timeout)
    except TimeoutError as exc:
        logger.debug(f"signal pva://{pv} timed out", exc_info=True)
        raise NotConnectedError(f"pva://{pv}") from exc


def _pva_request_string(fields: Sequence[str]) -> str:
    """Convert a list of requested fields into a PVA request string.

    This can be passed to p4p.
    """
    return f"field({','.join(fields)})"


class PvaSignalBackend(EpicsSignalBackend[SignalDatatypeT]):
    """Backend for a signal to interact with PVs over pva."""

    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_pv: str = "",
        write_pv: str = "",
        options: EpicsOptions | None = None,
    ):
        self.converter: PvaConverter = DisconnectedPvaConverter(float)
        self.initial_values: dict[str, Any] = {}
        self.subscription: Subscription | None = None
        super().__init__(datatype, read_pv, write_pv, options)

    def source(self, name: str, read: bool):
        return f"pva://{self.read_pv if read else self.write_pv}"

    async def _store_initial_value(self, pv: str, timeout: float):
        self.initial_values[pv] = await pvget_with_timeout(pv, timeout)

    async def connect(self, timeout: float):
        if self.read_pv != self.write_pv:
            # Different, need to connect both
            await wait_for_connection(
                read_pv=self._store_initial_value(self.read_pv, timeout=timeout),
                write_pv=self._store_initial_value(self.write_pv, timeout=timeout),
            )
        else:
            # The same, so only need to connect one
            await self._store_initial_value(self.read_pv, timeout=timeout)
        self.converter = make_converter(self.datatype, self.initial_values)

    def _make_reading(self, value: Any) -> Reading[SignalDatatypeT]:
        ts = value["timeStamp"]
        sv = value["alarm"]["severity"]
        return {
            "value": self.converter.value(value),
            "timestamp": ts["secondsPastEpoch"] + ts["nanoseconds"] * 1e-9,
            "alarm_severity": -1 if sv > 2 else sv,
        }

    async def put(self, value: SignalDatatypeT | None):
        if value is None:
            write_value = self.initial_values[self.write_pv]["value"]
        else:
            write_value = self.converter.write_value(value)
        if callable(self.options.wait):
            wait = self.options.wait(value)
        else:
            wait = self.options.wait
        await context().put(self.write_pv, {"value": write_value}, wait=wait)

    async def get_datakey(self, source: str) -> DataKey:
        value = await context().get(self.read_pv)
        metadata = _metadata_from_value(self.converter.datatype, value)
        return make_datakey(
            self.converter.datatype, self.converter.value(value), source, metadata
        )

    async def get_reading(self) -> Reading:
        request = _pva_request_string(
            self.converter.value_fields + self.converter.reading_fields
        )
        value = await context().get(self.read_pv, request=request)
        return self._make_reading(value)

    async def get_value(self) -> SignalDatatypeT:
        request = _pva_request_string(self.converter.value_fields)
        value = await context().get(self.read_pv, request=request)
        return self.converter.value(value)

    async def get_setpoint(self) -> SignalDatatypeT:
        request = _pva_request_string(self.converter.value_fields)
        value = await context().get(self.write_pv, request=request)
        return self.converter.value(value)

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        if callback and self.subscription:
            msg = "Cannot set a callback when one is already set"
            raise RuntimeError(msg)

        if self.subscription:
            self.subscription.close()
            self.subscription = None

        if callback:

            async def async_callback(v):
                callback(self._make_reading(v))

            request = _pva_request_string(
                self.converter.value_fields + self.converter.reading_fields
            )
            self.subscription = context().monitor(
                self.read_pv, async_callback, request=request
            )



import asyncio
import os
import subprocess
import time
from pathlib import Path

import h5py
import pytest

from ophyd_async.core import (
    DetectorTrigger,
    Device,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
)
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.fastcs.eiger import EigerDetector

SAVE_PATH = "/tmp"

_eiger_ioc = "EIGER-CI"
_odin_ioc = "ODIN-CI"
_eiger_ioc_image = "ghcr.io/diamondlightsource/eiger-fastcs:0.1.0beta5"
_odin_ioc_image = "ghcr.io/diamondlightsource/odin-fastcs:0.2.0beta2"

env = os.environ.copy()
env["eiger_ioc"] = _eiger_ioc
env["odin_ioc"] = _odin_ioc


class SetupDevice(Device):
    """Holds PVs that we would either expect to be initially set and
    never change or to be externally set in prod.
    """

    def __init__(self, eiger_prefix: str, odin_prefix: str) -> None:
        self.trigger = epics_signal_rw(int, f"{eiger_prefix}Trigger")
        self.header_detail = epics_signal_rw(str, f"{eiger_prefix}HeaderDetail")
        self.compression = epics_signal_rw(str, f"{odin_prefix}DatasetDataCompression")
        self.frames_per_block = epics_signal_rw(
            int, f"{odin_prefix}ProcessFramesPerBlock"
        )
        self.blocks_per_file = epics_signal_rw(
            int, f"{odin_prefix}ProcessBlocksPerFile"
        )
        super().__init__("")


@pytest.fixture(scope="session")
def eiger_ioc():
    # Start the IOC process
    command = (
        f"docker run --rm --name={_eiger_ioc} -dt --net=host -v /tmp/opi/:/epics/opi"
        f" {_eiger_ioc_image} ioc {_eiger_ioc}"
    )
    _ = subprocess.Popen(
        command.split(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )
    try:
        # wait for IOC to be ready (sleep or poll)
        time.sleep(1)
        yield
    finally:
        # Cleanup: terminate the IOC process
        os.system(f"docker kill {_eiger_ioc}")


@pytest.fixture(scope="session")
def odin_ioc():
    # Start the IOC process
    command = (
        f"docker run --rm --name={_odin_ioc} -dt --net=host -v /tmp/opi/:/epics/opi"
        f" {_odin_ioc_image} ioc {_odin_ioc}"
    )
    _ = subprocess.Popen(
        command.split(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )
    try:
        # wait for IOC to be ready (sleep or poll)
        time.sleep(1)
        yield
    finally:
        # Cleanup: terminate the IOC process
        os.system(f"docker kill {_odin_ioc}")


@pytest.fixture
def ioc_prefixes(eiger_ioc, odin_ioc):
    return _eiger_ioc + ":", _odin_ioc + ":"


@pytest.fixture
async def setup_device(RE, ioc_prefixes):
    async with init_devices():
        device = SetupDevice(ioc_prefixes[0], ioc_prefixes[1] + "FP:")
    await asyncio.gather(
        device.header_detail.set("all"),
        device.compression.set("BSLZ4"),
        device.frames_per_block.set(1000),
        device.blocks_per_file.set(1),
    )

    return device


@pytest.fixture
async def test_eiger(RE, ioc_prefixes) -> EigerDetector:
    provider = StaticPathProvider(StaticFilenameProvider("test_eiger"), Path(SAVE_PATH))
    async with init_devices():
        test_eiger = EigerDetector("", provider, ioc_prefixes[0], ioc_prefixes[1])

    return test_eiger


@pytest.mark.skip(reason="Flaky test, see GH-819")
async def test_trigger_saves_file(test_eiger: EigerDetector, setup_device: SetupDevice):
    single_shot = TriggerInfo(
        exposure_timeout=None,
        number_of_events=1,
        trigger=DetectorTrigger.INTERNAL,
        livetime=None,
    )

    await test_eiger.stage()
    await test_eiger.prepare(single_shot)
    # Need to work out what the hold up is in prepare so we cant do this straight away.
    # File path propogation?
    await asyncio.sleep(0.5)
    await setup_device.trigger.set(1)
    await asyncio.sleep(0.5)  # Need to work out when it's actually finished writing
    await test_eiger.unstage()

    with h5py.File(SAVE_PATH + "/test_eiger_000001.h5") as f:
        assert "data" in f.keys()



from ophyd_async.core import DetectorTriggerLogic, SignalDict

from ._block import PcapBlock


class PandaTriggerLogic(DetectorTriggerLogic):
    """For controlling a PCAP capture on the PandA."""

    def __init__(self, pcap: PcapBlock) -> None:
        self.pcap = pcap

    def get_deadtime(self, config_values: SignalDict) -> float:
        # Need 1 tick of the 125MHz clock as deadtime
        # https://quantumdetectors.com/products/beamline-data-acquisition-tool/
        return 8e-9

    async def prepare_level(self, num: int):
        pass



"""Helper script to test transport unavailability by blocking module imports."""

if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser(
        description="Test EPICS signal creation with blocked transports"
    )
    parser.add_argument(
        "--block-ca",
        action="store_true",
        help="Block aioca import to simulate CA unavailable",
    )
    parser.add_argument(
        "--block-pva",
        action="store_true",
        help="Block p4p import to simulate PVA unavailable",
    )
    parser.add_argument("pv_name", type=str, help="PV name to use for signal creation")

    args = parser.parse_args()

    # Block the requested transports
    if args.block_ca:
        sys.modules["aioca"] = None  # type: ignore

    if args.block_pva:
        sys.modules["p4p"] = None  # type: ignore
        sys.modules["p4p.client"] = None  # type: ignore
        sys.modules["p4p.client.thread"] = None  # type: ignore

    # Now try to import and use ophyd_async
    from ophyd_async.epics.core import epics_signal_rw

    # Try to create a signal - let exceptions bubble up naturally
    signal = epics_signal_rw(int, args.pv_name, name="test")



from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Generic, TypeVar, get_args, get_origin

import numpy as np

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    SignalBackend,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
    get_dtype,
    get_enum_cls,
    observe_value,
    wait_for_value,
)

T = TypeVar("T")


@dataclass
class EpicsOptions(Generic[SignalDatatypeT]):
    """Options for EPICS Signals."""

    wait: bool | Callable[[SignalDatatypeT], bool] = True
    """Whether to wait for server-side completion of the operation:

    - `True`: Return when server-side operation has completed
    - `False`: Return when server-side operation has started
    - `callable`: Call with the value being put to decide whether to wait

    For example, use `EpicsOption(wait=non_zero)` for busy records like
    areaDetector acquire PVs that should not wait when being set to zero
    as it causes a deadlock.
    """


def get_pv_basename_and_field(pv: str) -> tuple[str, str | None]:
    """Split PV into record name and field."""
    if "." in pv:
        record, field = pv.split(".", maxsplit=1)
    else:
        record, field = pv, None
    return (record, field)


def get_supported_values(
    pv: str,
    datatype: type[T],
    pv_choices: Sequence[str],
) -> Mapping[str, T | str]:
    enum_cls = get_enum_cls(datatype)
    if not enum_cls:
        raise TypeError(f"{datatype} is not an Enum")
    choices = [v.value for v in enum_cls]
    error_msg = f"{pv} has choices {pv_choices}, but {datatype} requested {choices} "
    if issubclass(enum_cls, StrictEnum):
        if set(choices) != set(pv_choices):
            raise TypeError(error_msg + "to be strictly equal to them.")
    elif issubclass(enum_cls, SubsetEnum):
        if not set(choices).issubset(pv_choices):
            raise TypeError(error_msg + "to be a subset of them.")
    elif issubclass(enum_cls, SupersetEnum):
        if not set(pv_choices).issubset(choices):
            raise TypeError(error_msg + "to be a superset of them.")
    else:
        raise TypeError(f"{datatype} is not a StrictEnum, SubsetEnum, or SupersetEnum")
    # Create a map from the string value to the enum instance
    # For StrictEnum and SupersetEnum, all values here will be enum values
    # For SubsetEnum, only the values in choices will be enum values, the rest will be
    # strings
    supported_values = {x: enum_cls(x) for x in pv_choices}
    return supported_values


def format_datatype(datatype: Any) -> str:
    if get_origin(datatype) is np.ndarray and get_args(datatype):
        dtype = get_dtype(datatype)
        return f"Array1D[np.{dtype.name}]"
    elif get_origin(datatype) is Sequence:
        return f"Sequence[{get_args(datatype)[0].__name__}]"
    elif isinstance(datatype, type):
        return datatype.__name__
    else:
        return str(datatype)


class EpicsSignalBackend(SignalBackend[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_pv: str = "",
        write_pv: str = "",
        options: EpicsOptions | None = None,
    ):
        self.read_pv = read_pv
        self.write_pv = write_pv
        self.options = options or EpicsOptions()
        super().__init__(datatype)


async def stop_busy_record(
    signal: SignalRW[bool],
    timeout: float = DEFAULT_TIMEOUT,
) -> None:
    await signal.set(False)
    await wait_for_value(signal, False, timeout=timeout)


async def wait_for_good_state(
    state_signal: SignalR[SignalDatatypeT],
    good_states: set[SignalDatatypeT],
    message_signal: SignalR[str] | None = None,
    timeout: float = DEFAULT_TIMEOUT,
) -> None:
    """Wait for state_signal to be one of good_states within timeout."""
    state: SignalDatatypeT | None = None
    try:
        async for state in observe_value(state_signal, done_timeout=timeout):
            if state in good_states:
                return
    except TimeoutError as exc:
        if state is None:
            # No updates from the detector, maybe it is disconnected
            raise TimeoutError(
                f"Could not monitor state: {state_signal.source}"
            ) from exc
        else:
            # No good state in time, report message if given
            msg = f"{state_signal.source} not in a good state: {state}: "
            if message_signal:
                msg += await message_signal.get_value()
            else:
                msg += f"expected one of {good_states}"
            raise ValueError(msg) from exc



"""Support for the SimDetector areaDetector driver.

https://github.com/areaDetector/ADSimDetector.
"""

from collections.abc import Sequence

from ophyd_async.core import (
    DetectorTriggerLogic,
    PathProvider,
    SignalR,
)

from .adcore import (
    ADArmLogic,
    ADBaseIO,
    ADWriterType,
    AreaDetector,
    NDPluginBaseIO,
    prepare_exposures,
)

__all__ = [
    "SimDetector",
    "SimDetectorTriggerLogic",
]


class SimDetectorTriggerLogic(DetectorTriggerLogic):
    """Trigger logic for ADSimDetector."""

    def __init__(self, driver: ADBaseIO):
        self.driver = driver

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await prepare_exposures(self.driver, num, livetime, deadtime)


class SimDetector(AreaDetector[ADBaseIO]):
    """Create an ADSimDetector AreaDetector instance.

    :param prefix: EPICS PV prefix for the detector
    :param path_provider: Provider for file paths during acquisition
    :param driver_suffix: Suffix for the driver PV, defaults to "cam1:"
    :param writer_type: Type of file writer (HDF or TIFF)
    :param writer_suffix: Suffix for the writer PV
    :param plugins: Additional areaDetector plugins to include
    :param config_sigs: Additional signals to include in configuration
    :param name: Name for the detector device
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        driver_suffix="cam1:",
        writer_type: ADWriterType | None = ADWriterType.HDF,
        writer_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        driver = ADBaseIO(prefix + driver_suffix)
        super().__init__(
            prefix=prefix,
            driver=driver,
            arm_logic=ADArmLogic(driver),
            trigger_logic=SimDetectorTriggerLogic(driver),
            path_provider=path_provider,
            writer_type=writer_type,
            writer_suffix=writer_suffix,
            plugins=plugins,
            config_sigs=config_sigs,
            name=name,
        )



"""Utilities for testing devices."""

from . import __pytest_assert_rewrite  # noqa: F401
from ._assert import (
    ApproxTable,
    MonitorQueue,
    StatusWatcher,
    approx_value,
    assert_configuration,
    assert_describe_signal,
    assert_emitted,
    assert_has_calls,
    assert_reading,
    assert_value,
    partial_reading,
)
from ._one_of_everything import (
    ExampleEnum,
    ExampleTable,
    OneOfEverythingDevice,
    ParentOfEverythingDevice,
    float_array_value,
    int_array_value,
)
from ._single_derived import (
    BeamstopPosition,
    Exploder,
    MovableBeamstop,
    ReadOnlyBeamstop,
)
from ._wait_for_pending import wait_for_pending_wakeups


# Back compat - delete before 1.0
def __getattr__(name):
    import warnings

    import ophyd_async.core

    moved_to_core = {
        "callback_on_mock_put",
        "get_mock",
        "get_mock_put",
        "mock_puts_blocked",
        "set_mock_put_proceeds",
        "set_mock_value",
        "set_mock_values",
    }
    if name in moved_to_core:
        warnings.warn(
            DeprecationWarning(
                f"ophyd_async.testing.{name} has moved to ophyd_async.core"
            ),
            stacklevel=2,
        )
        return getattr(ophyd_async.core, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


# The order of this list determines the order of the documentation,
# so does not match the alphabetical order of the imports
__all__ = [
    "approx_value",
    # Assert functions
    "assert_value",
    "assert_reading",
    "assert_configuration",
    "assert_describe_signal",
    "assert_emitted",
    "partial_reading",
    # Wait for pending wakeups
    "wait_for_pending_wakeups",
    "ExampleEnum",
    "ExampleTable",
    "OneOfEverythingDevice",
    "ParentOfEverythingDevice",
    "MonitorQueue",
    "ApproxTable",
    "StatusWatcher",
    "int_array_value",
    "float_array_value",
    # Derived examples
    "BeamstopPosition",
    "Exploder",
    "MovableBeamstop",
    "ReadOnlyBeamstop",
    "assert_has_calls",
]



from collections.abc import Awaitable

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator, plan

from ._utils import T


@plan
def wait_for_awaitable(coro: Awaitable[T]) -> MsgGenerator[T]:
    """Wait for a single awaitable to complete, and return the result."""
    (task,) = yield from bps.wait_for([lambda: coro])
    return task.result()



"""Support for the ADAndor areaDetector driver.

https://github.com/areaDetector/ADAndor.
"""

from collections.abc import Sequence
from typing import Annotated as A

from ophyd_async.core import (
    DetectorTriggerLogic,
    PathProvider,
    SignalDict,
    SignalR,
    SignalRW,
    StrictEnum,
)
from ophyd_async.epics.adcore import (
    ADArmLogic,
    ADBaseIO,
    ADWriterType,
    AreaDetector,
    NDPluginBaseIO,
    prepare_exposures,
)
from ophyd_async.epics.core import PvSuffix

__all__ = [
    "AndorDetector",
    "Andor2DriverIO",
    "Andor2TriggerLogic",
    "Andor2TriggerMode",
]

_MIN_DEAD_TIME = 0.1
_MAX_NUM_IMAGE = 999_999


class Andor2TriggerMode(StrictEnum):
    """Trigger modes for ADAndor detector."""

    INTERNAL = "Internal"
    EXT_TRIGGER = "External"
    EXT_START = "External Start"
    EXT_EXPOSURE = "External Exposure"
    EXT_FVP = "External FVP"
    SOFTWARE = "Software"


class Andor2DriverIO(ADBaseIO):
    """Driver for andor model:DU897_BV as deployed on p99.

    This mirrors the interface provided by AdAndor/db/andor.template.
    https://areadetector.github.io/areaDetector/ADAndor/andorDoc.html
    """

    trigger_mode: A[SignalRW[Andor2TriggerMode], PvSuffix.rbv("TriggerMode")]
    andor_accumulate_period: A[SignalR[float], PvSuffix("AndorAccumulatePeriod_RBV")]


# The deadtime of an Andor2 controller varies depending on the exact model of camera.
# Ideally we would maximize performance by dynamically retrieving the deadtime at
# runtime. See https://github.com/bluesky/ophyd-async/issues/308
class Andor2TriggerLogic(DetectorTriggerLogic):
    """Trigger logic for Andor2DriverIO."""

    def __init__(self, driver: Andor2DriverIO):
        self.driver = driver

    def get_deadtime(self, config_values: SignalDict) -> float:
        return _MIN_DEAD_TIME

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await self.driver.trigger_mode.set(Andor2TriggerMode.INTERNAL)
        await prepare_exposures(self.driver, num or _MAX_NUM_IMAGE, livetime, deadtime)

    async def prepare_edge(self, num: int, livetime: float):
        await self.driver.trigger_mode.set(Andor2TriggerMode.EXT_TRIGGER)
        await prepare_exposures(self.driver, num or _MAX_NUM_IMAGE, livetime)


class AndorDetector(AreaDetector[Andor2DriverIO]):
    """Create an ADAndor AreaDetector instance.

    :param prefix: EPICS PV prefix for the detector
    :param path_provider: Provider for file paths during acquisition
    :param driver_suffix: Suffix for the driver PV, defaults to "cam1:"
    :param writer_type: Type of file writer (HDF or TIFF)
    :param writer_suffix: Suffix for the writer PV
    :param plugins: Additional areaDetector plugins to include
    :param config_sigs: Additional signals to include in configuration
    :param name: Name for the detector device
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        driver_suffix="cam1:",
        writer_type: ADWriterType | None = ADWriterType.HDF,
        writer_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        driver = Andor2DriverIO(prefix + driver_suffix)
        super().__init__(
            prefix=prefix,
            driver=driver,
            arm_logic=ADArmLogic(driver),
            trigger_logic=Andor2TriggerLogic(driver),
            path_provider=path_provider,
            writer_type=writer_type,
            writer_suffix=writer_suffix,
            plugins=plugins,
            config_sigs=config_sigs,
            name=name,
        )



import re
from typing import Any

from ophyd_async.core import StrictEnum


class DevStateEnum(StrictEnum):
    ON = "ON"
    OFF = "OFF"
    CLOSE = "CLOSE"
    OPEN = "OPEN"
    INSERT = "INSERT"
    EXTRACT = "EXTRACT"
    MOVING = "MOVING"
    STANDBY = "STANDBY"
    FAULT = "FAULT"
    INIT = "INIT"
    RUNNING = "RUNNING"
    ALARM = "ALARM"
    DISABLE = "DISABLE"
    UNKNOWN = "UNKNOWN"


def get_full_attr_trl(device_trl: str, attr_name: str) -> str:
    device_parts = device_trl.split("#", 1)
    # my/device/name#dbase=no splits into my/device/name and
    # dbase=no
    full_trl = device_parts[0] + "/" + attr_name
    if len(device_parts) > 1:
        full_trl += "#" + device_parts[1]
    return full_trl


def get_device_trl_and_attr(name: str):
    # trl can have form:
    #   <protocol://><server:host/>domain/family/member/attr_name<#dbase=no>
    # e.g. tango://127.0.0.1:8888/test/nodb/test#dbase=no
    re_str = (
        r"([\.a-zA-Z0-9_-]*://)?([\.a-zA-Z0-9_-]+:[0-9]+/)?"
        r"([^#/]+/[^#/]+/[^#/]+/)([^#/]+)(#dbase=[a-z]+)?"
    )
    search = re.search(re_str, name)
    if not search:
        raise ValueError(f"Could not parse device and attribute from trl {name}")
    groups = [part if part else "" for part in search.groups()]
    attr = groups.pop(3)  # extract attr name from groups
    groups[2] = groups[2].removesuffix("/")  # remove trailing slash from device name
    device = "".join(groups)
    return device, attr


def try_to_cast_as_float(value: Any) -> float | None:
    """Attempt to cast a value to float, returning None on failure."""
    try:
        return float(value)
    except (ValueError, TypeError):
        return None
